# -*- coding: utf-8 -*-
"""
Module containing utilities regarding the usage of ogr functionalities.
"""

#-------------------------------------
# Import/init needed modules
#-------------------------------------
from concurrent import futures
from datetime import datetime
from io import StringIO
import logging
import multiprocessing
import os
from pathlib import Path
import pprint
import re
import subprocess
import tempfile
from threading import Lock
import time
from typing import List, Optional, Tuple, Union

import geopandas as gpd
from osgeo import gdal
gdal.UseExceptions() 

from geofileops import geofile
from geofileops.util import general_util
from geofileops.util.general_util import MissingRuntimeDependencyError
from geofileops.util.geometry_util import GeometryType

#-------------------------------------------------------------
# First define/init some general variables/constants
#-------------------------------------------------------------

# Make sure only one instance per process is running
lock = Lock()

# Get a logger...
logger = logging.getLogger(__name__)

# Cache the result of the check because it is quite expensive
global_spatiallite_versioninfo_default = None
global_spatiallite_versioninfo_alternative = None

#-------------------------------------------------------------
# The real work
#-------------------------------------------------------------

class VectorTranslateInfo:
    def __init__(
            self,
            input_path: Path, 
            output_path: Path,
            translate_description: str = None,
            input_layers: Union[Optional[List[str]], str] = None,
            output_layer: str = None,
            spatial_filter: Tuple[float, float, float, float] = None,
            clip_bounds: Tuple[float, float, float, float] = None, 
            sql_stmt: str = None,
            sql_dialect: str = None,
            transaction_size: int = 65536,
            append: bool = False,
            update: bool = False,
            create_spatial_index: bool = None,
            explodecollections: bool = False,
            force_output_geometrytype: GeometryType = None,
            priority_class: str = 'VERY_LOW',
            sqlite_journal_mode: str = None,
            force_py: bool = False,
            verbose: bool = False):
        self.input_path = input_path
        self.output_path = output_path
        self.translate_description = translate_description
        self.input_layers = input_layers
        self.output_layer = output_layer
        self.spatial_filter = spatial_filter
        self.clip_bounds = clip_bounds
        self.sql_stmt = sql_stmt
        self.sql_dialect = sql_dialect
        self.transaction_size = transaction_size
        self.append = append
        self.update = update
        self.create_spatial_index = create_spatial_index
        self.explodecollections = explodecollections
        self.force_output_geometrytype = force_output_geometrytype
        self.priority_class = priority_class
        self.sqlite_journal_mode = sqlite_journal_mode
        self.force_py = force_py
        self.verbose = verbose

def vector_translate_by_info(info: VectorTranslateInfo):

    return vector_translate( 
            input_path=info.input_path,
            output_path=info.output_path,
            translate_description=info.translate_description,
            input_layers=info.input_layers,
            output_layer=info.output_layer,
            spatial_filter=info.spatial_filter,
            clip_bounds=info.clip_bounds,
            sql_stmt=info.sql_stmt,
            sql_dialect=info.sql_dialect,
            transaction_size=info.transaction_size,
            append=info.append,
            update=info.update,
            create_spatial_index=info.create_spatial_index,
            explodecollections=info.explodecollections,
            force_output_geometrytype=info.force_output_geometrytype,
            priority_class=info.priority_class,    
            sqlite_journal_mode=info.sqlite_journal_mode,
            force_py=info.force_py,
            verbose=info.verbose)

def vector_translate_async(
        concurrent_pool,
        info: VectorTranslateInfo):

    return concurrent_pool.submit(
            vector_translate_by_info,
            info)

def vector_translate_seq(
        vector_translate_infos: List[VectorTranslateInfo]):

    for vector_translate_info in vector_translate_infos:
        vector_translate_by_info(vector_translate_info)

def vector_translate_parallel(
        vector_translate_infos: List[VectorTranslateInfo],
        operation_description: str,
        nb_parallel: int = -1):

    # Init
    if len(vector_translate_infos) == 0:
        return
    start_time = datetime.now()
    if nb_parallel == -1:
        nb_parallel = multiprocessing.cpu_count()
    
    with futures.ThreadPoolExecutor(nb_parallel) as concurrent_pool:
        
        future_to_calc_id = {}
        for calc_id, vector_translate_info in enumerate(vector_translate_infos):
            future = vector_translate_async(
                    concurrent_pool=concurrent_pool,
                    info=vector_translate_info)
            future_to_calc_id[future] = calc_id
        
        # Wait till all parallel processes are ready
        nb_done = 0
        for future in futures.as_completed(future_to_calc_id):
            calc_id = future_to_calc_id[future]
            try:
                _ = future.result()
            except Exception as ex:
                message = f"Async translate {calc_id} ERROR: {vector_translate_infos[calc_id].translate_description}\n{pprint.pformat(vector_translate_infos[calc_id])}"
                raise Exception(message) from ex

            # Log the progress and prediction speed
            nb_done += 1
            general_util.report_progress(start_time, nb_done, len(vector_translate_infos), operation_description)      

def vector_translate(
        input_path: Path, 
        output_path: Path,
        translate_description: str = None,
        input_layers: Union[Optional[List[str]], str] = None,
        output_layer: str = None,
        spatial_filter: Tuple[float, float, float, float] = None,
        clip_bounds: Tuple[float, float, float, float] = None, 
        sql_stmt: str = None,
        sql_dialect: str = None,
        transaction_size: int = 65536,
        append: bool = False,
        update: bool = False,
        create_spatial_index: bool = None,
        explodecollections: bool = False,
        force_output_geometrytype: GeometryType = None,
        priority_class: str = 'VERY_LOW',
        sqlite_journal_mode: str = None,
        force_py: bool = False,
        verbose: bool = False) -> bool:
    """
    Run a command
    """
    # If running a sqlite statement on a gpkg file, check first if spatialite works properly
    # Remark: if force_py is True, the user should know what he is doing 
    sql_dialect_to_use = sql_dialect
    if force_py:
        gdal_to_use = 'gdal_default'
    elif(sql_stmt is not None 
       and sql_dialect is not None
       and sql_dialect.upper() == 'SQLITE' 
       and input_path.suffix.lower() == '.gpkg'):
        gdal_to_use, sql_dialect_to_use = get_gdal_to_use(sql_stmt)
    else:
        gdal_to_use = 'gdal_default'

    if gdal_to_use == 'gdal_default':
        return vector_translate_py( 
            input_path=input_path,
            output_path=output_path,
            translate_description=translate_description,
            input_layers=input_layers,
            output_layer=output_layer,
            spatial_filter=spatial_filter,
            clip_bounds=clip_bounds,
            sql_stmt=sql_stmt,
            sql_dialect=sql_dialect_to_use,
            transaction_size=transaction_size,
            append=append,
            update=update,
            create_spatial_index=create_spatial_index,
            explodecollections=explodecollections,
            force_output_geometrytype=force_output_geometrytype,
            priority_class=priority_class,    
            sqlite_journal_mode=sqlite_journal_mode,
            verbose=verbose)
    elif gdal_to_use == 'gdal_bin':
        logger.debug("Use ogr2ogr.exe instead of python version")
        return vector_translate_exe( 
            input_path=input_path,
            output_path=output_path,
            translate_description=translate_description,
            input_layers=input_layers,
            output_layer=output_layer,
            spatial_filter=spatial_filter,
            clip_bounds=clip_bounds,
            sql_stmt=sql_stmt,
            sql_dialect=sql_dialect_to_use,
            transaction_size=transaction_size,
            append=append,
            update=update,
            create_spatial_index=create_spatial_index,
            explodecollections=explodecollections,
            force_output_geometrytype=force_output_geometrytype,
            priority_class=priority_class,    
            sqlite_journal_mode=sqlite_journal_mode,
            verbose=verbose)
    else:
        raise Exception(f"Unsupported gdal_to_use: {gdal_to_use}")

def vector_translate_exe(
        input_path: Path, 
        output_path: Path,
        translate_description: str = None,
        input_layers: Union[Optional[List[str]], str] = None,
        output_layer: str = None,
        spatial_filter: Tuple[float, float, float, float] = None,
        clip_bounds: Tuple[float, float, float, float] = None, 
        sql_stmt: str = None,
        sql_dialect: str = None,
        transaction_size: int = 10000,
        append: bool = False,
        update: bool = False,
        create_spatial_index: bool = None,
        explodecollections: bool = False,
        force_output_geometrytype: GeometryType = None,
        priority_class: str = 'VERY_LOW',
        sqlite_journal_mode: str = None,
        verbose: bool = False) -> bool:

    ##### Init #####
    if output_layer is None:
        output_layer = output_path.stem
    if os.name == 'nt':
        ogr2ogr_exe = 'ogr2ogr.exe'
    else:
        ogr2ogr_exe = 'ogr2ogr'
    
    # If GDAL_BIN is set, use ogr2ogr located there
    gdal_bin_dir = os.getenv('GDAL_BIN')
    if gdal_bin_dir is not None:
        gdal_bin_dir = Path(gdal_bin_dir)
        ogr2ogr_path = gdal_bin_dir / ogr2ogr_exe
    else:
        ogr2ogr_path = ogr2ogr_exe

    # Add all parameters to args list
    args = [str(ogr2ogr_path)]
    #if verbose:
    #    args.append('-progress')

    # Sql'ing, Filtering, clipping  
    if spatial_filter is not None:
        args.extend(['-spat', str(spatial_filter[0]), str(spatial_filter[1]), 
                    str(spatial_filter[2]), str(spatial_filter[3])])
    if clip_bounds is not None:
        args.extend(['-clipsrc', str(clip_bounds[0]), str(clip_bounds[1]), 
                    str(clip_bounds[2]), str(clip_bounds[3])])
    if sql_stmt is not None:
        args.extend(['-sql', sql_stmt])
    if sql_dialect is not None:
        args.extend(['-dialect', sql_dialect])

    # Output file options
    if output_path.exists() is True:
        if append is True:
            args.append('-append')
        if update is True:
            args.append('-update')

    # Files and input layer(s)
    args.append(str(output_path))
    args.append(str(input_path))
    
    # Treat input_layers variable
    # Remarks: 
    #   * For shapefiles, having input_layers not None gives issues
    #   * If a sql statement is passed, the input layers are not relevant,
    #     and ogr2ogr will give a warning, so clear it.
    if(input_layers is not None 
       and input_path.suffix.lower() != '.shp'
       and sql_stmt is None):
        if isinstance(input_layers, str):
            args.append(input_layers)
        else:
            for input_layer in input_layers:
                args.append(input_layer)
            
    # Output layer options
    if explodecollections is True:
        args.append('-explodecollections')
    if output_layer is not None:
        args.extend(['-nln', output_layer])
    if force_output_geometrytype is not None:
        args.extend(['-nlt', force_output_geometrytype.name])
    args.extend(['-nlt', 'PROMOTE_TO_MULTI'])
    if create_spatial_index is not None:
        if create_spatial_index is True:
            args.extend(['-lco', 'SPATIAL_INDEX=YES'])
        else:
            args.extend(['-lco', 'SPATIAL_INDEX=NO'])
    if transaction_size is not None:
        args.extend(['-gt', str(transaction_size)])
    
    # Sqlite specific options:
    output_filetype = geofile.GeofileType(output_path)
    if output_filetype == geofile.GeofileType.SQLite:
        # Use the spatialite type of sqlite
        args.extend(['-dsco', 'SPATIALITE=YES'])

    '''
    # Try if the busy_timeout isn't giving problems rather than solving them...
    if sqlite_journal_mode is not None:
        args.extend(['--config', 'OGR_SQLITE_PRAGMA', f"journal_mode={sqlite_journal_mode},busy_timeout=5000"])  
    else:
        args.extend(['--config', 'OGR_SQLITE_PRAGMA', 'busy_timeout=5000'])  
    '''
    if sqlite_journal_mode is not None:
        args.extend(['--config', 'OGR_SQLITE_PRAGMA', f"journal_mode={sqlite_journal_mode}"])  

    #if append is False:
    #    args.extend(['--config', 'OGR_SQLITE_PRAGMA', '"journal_mode=WAL"'])
    #    args.extend(['-dsco', 'ADD_GPKG_OGR_CONTENTS=NO'])
    #else:
    #    args.extend(['--config', 'OGR_SQLITE_PRAGMA', 'busy_timeout=-1'])  
    #args.extend(['--config', 'OGR_SQLITE_SYNCHRONOUS', 'OFF'])  
    args.extend(['--config', 'OGR_SQLITE_CACHE', '128'])

    ##### Now start ogr2ogr #####
    # Save whether the output file exists already prior to the operation
    output_path_exists_already = output_path.exists()

    # Make sure the output dir exists
    os.makedirs(output_path.parent, exist_ok=True)

    # Set priority of the process on Windows, so computer stays responsive
    if os.name == 'nt':  
        if priority_class is None or priority_class == 'NORMAL':
            creationflags = 0x00000020 # = NORMAL_PRIORITY_CLASS
        elif priority_class == 'VERY_LOW':
            creationflags = 0x00000040 # =IDLE_PRIORITY_CLASS
        elif priority_class == 'LOW':
            creationflags = 0x00004000 # =BELOW_NORMAL_PRIORITY_CLASS
        else: 
            raise Exception("Unsupported priority class: {priority_class}, use one of: 'NORMAL', 'LOW' or 'VERY_LOW'")
    else:
        # On non-windows os'es, creationflags cannot be used
        creationflags = 0

    # Geopackage/sqlite files are very sensitive for being locked, so retry till 
    # file is not locked anymore... 
    # TODO: ideally, the child processes would die when the parent is killed!
    returncode = None
    err = None
    for retry_count in range(10):
        if translate_description is not None:
            logger.debug(f"Start '{translate_description}' with retry_count: {retry_count}")

        # Start process!
        process = subprocess.Popen(args, 
                creationflags=creationflags,
                stdout=subprocess.PIPE, stderr=subprocess.PIPE, bufsize=-1, encoding='utf-8')
        output, err = process.communicate()
        returncode = process.returncode

        # If an error occured
        if returncode > 0:
            if str(err).startswith("ERROR 1: database is locked"):
                logger.warn(f"'ERROR 1: database is locked' occured during {translate_description}, retry nb: {retry_count}")
                time.sleep(1)
                continue
            else:
                # If output_path didn't exist yet before, clean it up... if it exists
                if not output_path_exists_already:
                    if output_path.exists():
                        geofile.remove(output_path)
                raise Exception(f"Error executing {pprint.pformat(args)}\n\t-> Return code: {returncode}\n\t-> Error: {err}\n\t->Output: {output}")
        elif(err is not None and err != ""
             and not str(err).startswith(r"Warning 1: Layer creation options ignored since an existing layer is")):
            # No error, but data (warnings) in stderr
            # Remark: the warning about layer creating options is common and not interesting: ignore!
            logger.warn(f"Finished executing {pprint.pformat(args)}")
            logger.warn(f"\t-> Returncode ok, but stderr contains: {err}")

        # Check if the output file contains data
        if output_path.exists():
            fileinfo = _getfileinfo(output_path, readonly=False)
            if len(fileinfo['layers']) == 0:
                output_path.unlink()
                if verbose is True:
                    logger.warn(f"Finished, but empty result for '{translate_description}'")
            elif translate_description is not None:
                if verbose is True:
                    logger.info(f"Finished '{translate_description}'")
                else:
                    logger.debug(f"Finished '{translate_description}'")
        else:
            if verbose is True:
                logger.warn(f"Finished, but empty result for '{translate_description}'")
                
        return True

    # If we get here, the retries didn't suffice to get it executed properly
    raise Exception(f"Error executing {pprint.pformat(args)}\n\t-> Return code: {returncode}\n\t-> Error: {err}")

def vector_translate_py(
        input_path: Path, 
        output_path: Path,
        translate_description: str = None,
        input_layers: Union[Optional[List[str]], str] = None,
        output_layer: str = None,
        spatial_filter: Tuple[float, float, float, float] = None,
        clip_bounds: Tuple[float, float, float, float] = None, 
        sql_stmt: str = None,
        sql_dialect: str = None,
        transaction_size: int = 65536,
        append: bool = False,
        update: bool = False,
        create_spatial_index: bool = None,
        explodecollections: bool = False,
        force_output_geometrytype: GeometryType = None,
        priority_class: str = 'VERY_LOW',
        sqlite_journal_mode: str = None,
        verbose: bool = False) -> bool:

    # Remark: when executing a select statement, I keep getting error that 
    # there are two columns named "geom" as he doesnt see the "geom" column  
    # in the select as a geometry column. Probably a version issue. Maybe 
    # try again later.
    args = []

    # Cleanup the input_layers variable.
    if input_path.suffix.lower() == '.shp':
        # For shapefiles, having input_layers not None gives issues
        input_layers = None
    elif sql_stmt is not None:
        # If a sql statement is passed, the input layers are not relevant,
        # and ogr2ogr will give a warning, so clear it.
        input_layers = None
     
    # Sql'ing, Filtering, clipping  
    if spatial_filter is not None:
        args.extend(['-spat', str(spatial_filter[0]), str(spatial_filter[1]), 
                    str(spatial_filter[2]), str(spatial_filter[3])])
    if clip_bounds is not None:
        args.extend(['-clipsrc', str(clip_bounds[0]), str(clip_bounds[1]), 
                    str(clip_bounds[2]), str(clip_bounds[3])])
    '''
    if sqlite_stmt is not None:
        args.extend(['-sql', sqlite_stmt, '-dialect', 'sqlite'])
    '''

    # Output file options
    if output_path.exists() is True:
        if append is True:
            args.append('-append')
        if update is True:
            args.append('-update')

    # Files
    #args.append(output_path)
    #args.append(input_path)

    # Output layer options
    if explodecollections is True:
        args.append('-explodecollections')
    if output_layer is not None:
        args.extend(['-nln', output_layer])
    if force_output_geometrytype is not None:
        args.extend(['-nlt', force_output_geometrytype.name])
    args.extend(['-nlt', 'PROMOTE_TO_MULTI'])
    if transaction_size is not None:
        args.extend(['-gt', str(transaction_size)])

    # Output layer creation options
    layerCreationOptions = []
    # TODO: should check if the layer exists instead of the file
    if not output_path.exists():
        if create_spatial_index is not None:
            if create_spatial_index is True:
                layerCreationOptions.extend(['SPATIAL_INDEX=YES'])
            else:
                layerCreationOptions.extend(['SPATIAL_INDEX=NO'])
    
    # Get output format from the filename
    output_filetype = geofile.GeofileType(output_path)

    # Sqlite specific options
    datasetCreationOptions = []
    if output_filetype == geofile.GeofileType.SQLite:
        # Use the spatialite type of sqlite
        #datasetCreationOptions.extend(['-dsco', 'SPATIALITE=YES'])
        datasetCreationOptions.append('SPATIALITE=YES')
      
    '''
    # Try if the busy_timeout isn't giving problems rather than solving them...
    if sqlite_journal_mode is not None:
        datasetCreationOptions.extend(['--config', 'OGR_SQLITE_PRAGMA', f"journal_mode={sqlite_journal_mode},busy_timeout=5000"])  
    else:
        datasetCreationOptions.extend(['--config OGR_SQLITE_PRAGMA busy_timeout=5000'])  
    '''
    if sqlite_journal_mode is not None:
        gdal.SetConfigOption('OGR_SQLITE_PRAGMA', f"journal_mode={sqlite_journal_mode}")

    #if append is False:
    #    args.extend(['--config', 'OGR_SQLITE_PRAGMA', '"journal_mode=WAL"'])
    #    args.extend(['-dsco', 'ADD_GPKG_OGR_CONTENTS=NO'])
    #else:
    #    args.extend(['--config', 'OGR_SQLITE_PRAGMA', 'busy_timeout=-1'])  
    #args.extend(['--config', 'OGR_SQLITE_SYNCHRONOUS', 'OFF'])  
    gdal.SetConfigOption('OGR_SQLITE_CACHE', '128')

    options = gdal.VectorTranslateOptions(
            options=args, 
            format=output_filetype.ogrdriver, 
            accessMode=None, 
            srcSRS=None, 
            dstSRS=None, 
            reproject=False, 
            SQLStatement=sql_stmt,
            SQLDialect=sql_dialect,
            where=None, #"geom IS NOT NULL", 
            selectFields=None, 
            addFields=False, 
            forceNullable=False, 
            spatFilter=spatial_filter, 
            spatSRS=None,
            datasetCreationOptions=datasetCreationOptions, 
            layerCreationOptions=layerCreationOptions, 
            layers=input_layers,
            layerName=output_layer,
            geometryType=None, 
            dim=None, 
            segmentizeMaxDist=None, 
            zField=None, 
            skipFailures=False, 
            limit=None, 
            callback=None, 
            callback_data=None)

    input_ds = None
    try: 
        # In some cases gdal only raises the last exception instead of the stack in VectorTranslate, 
        # so you lose necessary details! -> uncomment gdal.DontUseExceptions() when debugging!
        #gdal.DontUseExceptions()
        logger.debug(f"Execute {sql_stmt} on {input_path}")
        input_ds = gdal.OpenEx(str(input_path))

        # TODO: memory output support might be interesting to support
        result_ds = gdal.VectorTranslate(
                destNameOrDestDS=str(output_path),
                srcDS=input_ds,
                #SQLStatement=sql_stmt,
                #SQLDialect=sql_dialect,
                #layerName=output_layer
                options=options)
        if result_ds is None:
            raise Exception("BOEM")
        else:
            if result_ds.GetLayerCount() == 0:
                del result_ds
                if output_path.exists():
                    geofile.remove(output_path)
    except Exception as ex:
        message = f"Error executing {sql_stmt}"
        logger.exception(message)
        raise Exception(message) from ex
    finally:
        if input_ds is not None:
            del input_ds
        
    return True

def _getfileinfo(
        path: Path,
        readonly: bool = True,
        verbose: bool = False) -> dict:
            
    # Get info
    info_str = vector_info(
            path=path, 
            readonly=readonly,
            verbose=verbose)

    # Prepare result
    result_dict = {}
    result_dict['info_str'] = info_str
    result_dict['layers'] = []
    info_strio = StringIO(str(info_str))
    for line in info_strio.readlines():
        line = line.strip()
        if re.match(r"\A\d: ", line):
            # It is a layer, now extract only the layer name
            logger.debug(f"This is a layer: {line}")
            layername_with_geomtype = re.sub(r"\A\d: ", "", line)
            layername = re.sub(r" [(][a-zA-Z ]+[)]\Z", "", layername_with_geomtype)
            result_dict['layers'].append(layername)

    return result_dict

'''
def vector_info(
        path: Path, 
        task_description = None,
        layer: str = None,
        readonly: bool = False,
        report_summary: bool = False,
        sql_stmt: str = None,
        sql_dialect: str = None, 
        skip_health_check: bool = False,
        force_py: bool = False,      
        verbose: bool = False):
    """Run a command"""

    ##### Init #####
    if not path.exists():
        raise Exception(f"File does not exist: {path}")

    sql_dialect_to_use = sql_dialect
    if force_py:
        gdal_to_use = 'gdal_default'
    elif(sql_stmt is not None 
       and sql_dialect is not None
       and sql_dialect.upper() == 'SQLITE' 
       and path.suffix.lower() == '.gpkg'):
        gdal_to_use, sql_dialect_to_use = get_gdal_to_use(sql_stmt)
    else:
        gdal_to_use = 'gdal_default'

    if gdal_to_use == 'gdal_default':
        vector_info_py(
                path=path, 
                task_description=task_description,
                layer=layer,
                readonly=readonly,
                report_summary=report_summary,
                sql_stmt=sql_stmt,
                sql_dialect=sql_dialect_to_use, 
                skip_health_check=skip_health_check,      
                verbose=verbose)
    elif gdal_to_use == 'gdal_bin':
        vector_info_exe(
                path=path, 
                task_description=task_description,
                layer=layer,
                readonly=readonly,
                report_summary=report_summary,
                sql_stmt=sql_stmt,
                sql_dialect=sql_dialect, 
                skip_health_check=skip_health_check,      
                verbose=verbose)
    else:
        raise Exception(f"Unsupported gdal_to_use: {gdal_to_use}")

def vector_info_py(
        path: Path, 
        task_description = None,
        layer: str = None,
        readonly: bool = False,
        report_summary: bool = False,
        sql_stmt: str = None,
        sql_dialect: str = None, 
        skip_health_check: bool = False,      
        verbose: bool = False):
    """"Run a command"""

    ##### Init #####
    if not path.exists():
        raise Exception(f"File does not exist: {path}")

    if(skip_health_check is False
       and sql_stmt is not None 
       and sql_dialect is not None
       and sql_dialect.upper() == 'SQLITE' 
       and path.suffix.lower() == '.gpkg'):
        _, sql_dialect = get_gdal_to_use(sql_stmt)

        # Warn that updates with 'INDIRECT_SQLITE' give terrible performance!
        if sql_dialect == 'INDIRECT_SQLITE':
            logger.warn(f"sql_stmt needs 'INDIRECT_SQLITE', but this is gonna be slow!\n\t{sql_stmt}")  

    # Add all parameters to args list
    args = []
    #args.extend(['--config', 'OGR_SQLITE_PRAGMA', 'journal_mode=WAL'])  
    if readonly is True:
        args.append('-ro')
    if report_summary is True:
        args.append('-so')
    if sql_stmt is not None:
        args.extend(['-sql', sql_stmt])
    if sql_dialect is not None:
        args.extend(['-dialect', sql_dialect])

    # File and optionally the layer
    args.append(str(path))
    if layer is not None:
        # ogrinfo doesn't like + need quoted layer names, so remove single and double quotes
        layer_stripped = layer.strip("'\"")
        args.append(layer_stripped)

    info_options = gdal.InfoOptions(
            options=args, 
            format='text', 
            deserialize=True, 
            computeMinMax=False, 
            reportHistograms=False, 
            reportProj4=False, 
            stats=False, 
            approxStats=False, 
            computeChecksum=False, 
            showGCPs=True, 
            showMetadata=True, 
            showRAT=True, 
            showColorTable=True, 
            listMDD=False, 
            showFileList=True, 
            allMetadata=False, 
            extraMDDomains=None, 
            wktFormat=None)

    result = gdal.Info(
            ds=str(path),
            options=info_options)
'''

def vector_info(
        path: Path, 
        task_description = None,
        layer: str = None,
        readonly: bool = False,
        report_summary: bool = False,
        sql_stmt: str = None,
        sql_dialect: str = None, 
        skip_health_check: bool = False,      
        verbose: bool = False):
    """"Run a command"""

    ##### Init #####
    if not path.exists():
        raise Exception(f"File does not exist: {path}")
    if os.name == 'nt':
        ogrinfo_exe = 'ogrinfo.exe'
    else:
        ogrinfo_exe = 'ogrinfo'
    
    # If GDAL_BIN is set, use ogrinfo located there
    gdal_bin_dir = os.getenv('GDAL_BIN')
    if gdal_bin_dir is not None:
        gdal_bin_dir = Path(gdal_bin_dir)
        ogrinfo_path = gdal_bin_dir / ogrinfo_exe
    else:
        ogrinfo_path = ogrinfo_exe

    # If running a sqlite statement on a gpkg file, check first if spatialite 
    # works properly an if INDIRECT_SQLITE should be used.
    if(skip_health_check is False
       and sql_stmt is not None 
       and sql_dialect is not None
       and sql_dialect.upper() == 'SQLITE' 
       and path.suffix.lower() == '.gpkg'):
        _, sql_dialect = get_gdal_to_use(sql_stmt)

        # Warn that updates with 'INDIRECT_SQLITE' give terrible performance!
        if sql_dialect == 'INDIRECT_SQLITE':
            logger.warn(f"sql_stmt needs 'INDIRECT_SQLITE', but this is gonna be slow!\n\t{sql_stmt}")  

    # Add all parameters to args list
    args = [str(ogrinfo_path)]
    #args.extend(['--config', 'OGR_SQLITE_PRAGMA', 'journal_mode=WAL'])  
    if readonly is True:
        args.append('-ro')
    if report_summary is True:
        args.append('-so')
    if sql_stmt is not None:
        args.extend(['-sql', sql_stmt])
    if sql_dialect is not None:
        args.extend(['-dialect', sql_dialect])

    # File and optionally the layer
    args.append(str(path))
    if layer is not None:
        # ogrinfo doesn't like + need quoted layer names, so remove single and double quotes
        layer_stripped = layer.strip("'\"")
        args.append(layer_stripped)

    # TODO: ideally, the child processes would die when the parent is killed!
    # Start on low priority, so computer stays responsive
    if os.name == 'nt':
        #BELOW_NORMAL_PRIORITY_CLASS = 0x00004000
        creationflags = 0x00000040
    else:
        # On linux, needs to be 0, otherwise error!
        creationflags = 0

    ##### Run ogrinfo #####
    # Geopackage/sqlite files are very sensitive for being locked, so retry till 
    # file is not locked anymore...
    sleep_time = 1
    returncode = None 
    for retry_count in range(10):
        process = subprocess.Popen(args, 
                creationflags=creationflags,
                stdout=subprocess.PIPE, stderr=subprocess.PIPE, bufsize=-1, encoding='utf-8')
        output, err = process.communicate()
        returncode = process.returncode

        # If an error occured
        if returncode > 0:
            if str(err).startswith("ERROR 1: database is locked"):
                logger.warn(f"'ERROR 1: database is locked' occured during {task_description}, retry nb: {retry_count}")
                time.sleep(sleep_time)
                sleep_time += 1
                continue
            else:
                raise Exception(f"Error executing {pprint.pformat(args)}\n\t-> Return code: {returncode}\n\t-> Error: {err}\n\t->Output: {output}")  
        elif err is not None and err != "":
            # ogrinfo apparently sometimes give a wrong returncode, so if data 
            # in stderr, treat as error as well
            raise Exception(f"Error executing {pprint.pformat(args)}\n\t->Return code: {returncode}\n\t->Error: {err}\n\t->Output: {output}")
        elif verbose is True:
            logger.info(f"Ready executing {pprint.pformat(args)}")

        return output

    # If we get here, the retries didn't suffice to get it executed properly
    raise Exception(f"Error executing {pprint.pformat(args)}\n\t-> Return code: {returncode}")

def get_gdal_to_use(sqlite_stmt: str) -> Tuple[str, str]:
    """
    Check which gdal installation to use for this query: 
      * If query is supported using standard python execution, returns 'gdal_default'
      * If query is only supported using gedal installed in GDAL_BIN, returns 'gdal_bin'
      * If query is not supported at all, raises exception

    Raises:
        SQLNotSupportedException: [description]
    """
    ### First check if the python version can be used ###
    # Get spatialite version info for the default gdal 
    global global_spatiallite_versioninfo_default
    if global_spatiallite_versioninfo_default is None:
        global_spatiallite_versioninfo_default = get_gdal_install_info('gdal_default')

    # If version 5 and rttopo installed... the default python way can be used!
    if(global_spatiallite_versioninfo_default['spatialite_version()'] >= '5.0' 
       and global_spatiallite_versioninfo_default['rttopo_version()'] is not None):
        return ('gdal_default', 'SQLITE')
        
    # Check if there is an unsupported function in the sqlite statement
    sqlite_stmt_lower = sqlite_stmt.lower()
    unsupported_function_found = False 
    for unsupported_function in global_spatiallite_versioninfo_default['unsupported_functions']:
        if(f"{unsupported_function}(" in sqlite_stmt_lower
           or f"{unsupported_function} (" in sqlite_stmt_lower):
            unsupported_function_found = True
            break

    # If no unsupported function found, default gdal can be used
    if unsupported_function_found is False:
        return ('gdal_default', 'SQLITE')
    else:
        # If no spatial index table is used in sql_stmt, we can force indirect sql 
        # TODO: could be checked in a more robust way...
        if not 'join "rtree_' in sqlite_stmt_lower:
            return ('gdal_default', 'INDIRECT_SQLITE')

    ### Default gdal not OK, so check if there is a GDAL_BIN version that is ok ###
    # First check if there is an alternative gdal installation 
    gdal_bin_dir = os.getenv('GDAL_BIN')
    if gdal_bin_dir is None:
        raise MissingRuntimeDependencyError('sqlite_stmt not supported by default gdal, and no alternative gdal installation specified using GDAL_BIN')

    # Get spatiallite version info for the alternative gdal, if it isn't cached yet for this gdal_bin_dir
    global global_spatiallite_versioninfo_alternative
    if global_spatiallite_versioninfo_alternative is None:
        global_spatiallite_versioninfo_alternative = get_gdal_install_info('gdal_bin')
        global_spatiallite_versioninfo_alternative['gdal_bin_dir'] = gdal_bin_dir
    elif global_spatiallite_versioninfo_alternative['gdal_bin_dir'] != gdal_bin_dir:
        global_spatiallite_versioninfo_alternative = get_gdal_install_info('gdal_bin')
        global_spatiallite_versioninfo_alternative['gdal_bin_dir'] = gdal_bin_dir

    # Check if there is an unsupported function in the sqlite statement
    sqlite_stmt_lower = sqlite_stmt.lower()
    unsupported_function_found = False 
    for unsupported_function in global_spatiallite_versioninfo_alternative['unsupported_functions']:
        if(f"{unsupported_function}(" in sqlite_stmt_lower
           or f"{unsupported_function} (" in sqlite_stmt_lower):
            unsupported_function_found = True
            break

    # If no unsupported function found, alternative gdal can be used
    if unsupported_function_found is False:
        return ('gdal_bin', 'SQLITE')
    else:
        raise MissingRuntimeDependencyError(f"Alternative gdal in gdal_bin_dir {gdal_bin_dir} doesn't support sqlite_stmt either: {sqlite_stmt}")

def get_gdal_install_info(gdal_installation: str) -> dict:

    install_info = {}
    
    # First check the spatialite version
    test_path = Path(__file__).resolve().parent / "test.gpkg"
    sqlite_stmt = f'select spatialite_version()'    
    install_info_gdf = _execute_sql(test_path, sqlite_stmt, gdal_installation, sql_dialect='SQLITE')
        
    # Now get extra information, depending on the spatialite version
    if install_info_gdf['spatialite_version()'].values[0] >= '5.0.0':
        sqlite_stmt = f'select spatialite_version(), HasGeos(), HasGeosAdvanced(), HasGeosTrunk(), geos_version(), rttopo_version()'
    elif install_info_gdf['spatialite_version()'].values[0] >= '4.3.0':
        sqlite_stmt = f'select spatialite_version(), HasGeos(), HasGeosAdvanced(), HasGeosTrunk(), geos_version(), lwgeom_version()'
    else:
        raise Exception(f"Unsupported spatialite version: {install_info_gdf['spatialite_version()'][0]}")
    install_info_gdf = _execute_sql(test_path, sqlite_stmt, gdal_installation, sql_dialect='SQLITE')
    
    # Copy results to result dict
    for column in install_info_gdf.columns:
        install_info[column] = install_info_gdf[column][0]

    # Check if there are unsupported functions
    if install_info['spatialite_version()'] >= '5.0.0':
        if install_info['rttopo_version()'] is None:
            install_info['unsupported_functions'] = [
                    'makevalid', 'st_makevalid', 'isvalid', 'st_isvalid', 
                    'isvalidreason', 'st_isvalidreason', 'isvaliddetail', 'st_isvaliddetail']
        else:
            install_info['unsupported_functions'] = []
    elif install_info['spatialite_version()'] >= '4.3.0':
        if install_info['lwgeom_version()'] is None:
            install_info['unsupported_functions'] = [
                    'buffer', 'st_buffer', 
                    'convexhull', 'st_convexhull', 'simplify', 'st_simplify', 
                    'simplifypreservetopology', 'st_simplifypreservetopology', 
                    'makevalid', 'st_makevalid', 'isvalid', 'st_isvalid', 
                    'isvalidreason', 'st_isvalidreason', 'isvaliddetail', 'st_isvaliddetail',
                    'st_area', 'area', 'st_npoints', 'npoints',
                    'st_multi', 'collectionextract', 'st_intersection', 'st_union', 'st_collect']
        else:
            install_info['unsupported_functions'] = []

    return install_info

def _execute_sql(
        path: Path,
        sqlite_stmt: str,
        gdal_installation: str,
        sql_dialect: str) -> gpd.GeoDataFrame:
    
    with tempfile.TemporaryDirectory() as tmpdir:
        tmp_path = Path(tmpdir) / 'ogr_util_execute_sql_tmp_file.gpkg'
        if gdal_installation == 'gdal_default':
            vector_translate_py(
                    input_path=path,
                    output_path=tmp_path,
                    sql_stmt=sqlite_stmt,
                    sql_dialect=sql_dialect)
        elif gdal_installation == 'gdal_bin':
            vector_translate_exe(
                    input_path=path,
                    output_path=tmp_path,
                    sql_stmt=sqlite_stmt,
                    sql_dialect=sql_dialect)
        else:
            raise Exception(f"Unsupported gdal_installation: {gdal_installation}")
        
        # Return result
        install_info_gdf = geofile.read_file(tmp_path)
        return install_info_gdf

if __name__ == '__main__':
    None



# -*- coding: utf-8 -*-
"""
Tests for operations that are executed using a sql statement on two layers.
"""

from pathlib import Path
import sys

# Add path so the local geofileops packages are found 
sys.path.insert(0, str(Path(__file__).resolve().parent.parent))
from geofileops import geofile
from geofileops.geofile import GeometryType, PrimitiveType
from geofileops.util import geofileops_sql
from geofileops.util.general_util import MissingRuntimeDependencyError
from tests import test_helper

def test_erase_gpkg(tmpdir):
    # Erase from polygon layer, with and without gdal_bin set
    input_path = test_helper.TestFiles.polygons_parcels_gpkg
    erase_path = test_helper.TestFiles.polygons_zones_gpkg
    output_path = Path(tmpdir) / 'parcels-2020_erase_zones.gpkg'
    basetest_erase(input_path, erase_path, output_path, 
            expected_output_geometrytype=GeometryType.MULTIPOLYGON,
            gdal_installation='gdal_bin')
    basetest_erase(input_path, erase_path, output_path, 
            expected_output_geometrytype=GeometryType.MULTIPOLYGON,
            gdal_installation='gdal_default')

    # Erase from point layer, with and without gdal_bin set
    input_path = test_helper.TestFiles.points_gpkg
    erase_path = test_helper.TestFiles.polygons_zones_gpkg
    output_path = Path(tmpdir) / 'points_erase_zones.gpkg'
    basetest_erase(input_path, erase_path, output_path, 
            expected_output_geometrytype=GeometryType.MULTIPOINT,
            gdal_installation='gdal_default')
    basetest_erase(input_path, erase_path, output_path,
            expected_output_geometrytype=GeometryType.MULTIPOINT,
            gdal_installation='gdal_bin')

    # Erase from line layer, with and without gdal_bin set
    input_path = test_helper.TestFiles.linestrings_rows_of_trees_gpkg
    erase_path = test_helper.TestFiles.polygons_zones_gpkg
    output_path = Path(tmpdir) / 'rows_of_trees_erase_zones.gpkg'
    basetest_erase(input_path, erase_path, output_path, 
            expected_output_geometrytype=GeometryType.MULTILINESTRING,
            gdal_installation='gdal_default')
    basetest_erase(input_path, erase_path, output_path, 
            expected_output_geometrytype=GeometryType.MULTILINESTRING,
            gdal_installation='gdal_bin')

def test_erase_shp(tmpdir):
    # Prepare input and output paths
    input_path = test_helper.TestFiles.polygons_parcels_shp
    erase_path = test_helper.TestFiles.polygons_zones_gpkg
    output_path = Path(tmpdir) / 'parcels-2020_erase_zones.shp'

    # Try both with and without gdal_bin set
    basetest_erase(input_path, erase_path, output_path, 
            expected_output_geometrytype=GeometryType.MULTIPOLYGON,
            gdal_installation='gdal_bin')
    basetest_erase(input_path, erase_path, output_path, 
            expected_output_geometrytype=GeometryType.MULTIPOLYGON,
            gdal_installation='gdal_default')

def basetest_erase(
        input_path: Path,
        erase_path: Path, 
        output_basepath: Path,
        expected_output_geometrytype: GeometryType, 
        gdal_installation: str):

    # Do operation
    output_path = output_basepath.parent / f"{output_basepath.stem}_{gdal_installation}{output_basepath.suffix}"
    with test_helper.GdalBin(gdal_installation):
        ok_expected = test_helper.check_runtime_dependencies_ok('twolayer', gdal_installation)
        try:
            geofileops_sql.erase(
                    input_path=input_path, erase_path=erase_path,
                    output_path=output_path)
            test_ok = True
        except MissingRuntimeDependencyError:
            test_ok = False
    assert test_ok is ok_expected, f"Error: for {gdal_installation}, test_ok: {test_ok}, expected: {ok_expected}"

    # If it is expected not to be OK, don't do other checks
    if ok_expected is False:
        return

    # Now check if the tmp file is correctly created
    assert output_path.exists() == True
    layerinfo_orig = geofile.get_layerinfo(input_path)
    layerinfo_output = geofile.get_layerinfo(output_path)
    assert len(layerinfo_orig.columns) == len(layerinfo_output.columns)

    # Checks depending on geometry type
    assert layerinfo_output.geometrytype == expected_output_geometrytype
    if expected_output_geometrytype == GeometryType.MULTIPOLYGON:
        assert layerinfo_output.featurecount == 37
    elif expected_output_geometrytype == GeometryType.MULTIPOINT:
        assert layerinfo_output.featurecount == 47
    elif expected_output_geometrytype == GeometryType.MULTILINESTRING:
        assert layerinfo_output.featurecount == 12
    else:
        raise Exception(f"Unsupported expected_output_geometrytype: {expected_output_geometrytype}")

    # Now check the contents of the result file
    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None

def test_export_by_location_gpkg(tmpdir):
    # Prepare input and output paths
    input_to_select_from_path = test_helper.TestFiles.polygons_parcels_gpkg
    input_to_compare_with_path = test_helper.TestFiles.polygons_zones_gpkg
    output_path = Path(tmpdir) / 'parcels-2020_loc_zones.gpkg'

    # Try both with and without gdal_bin set
    basetest_export_by_location(
            input_to_select_from_path, input_to_compare_with_path, output_path, 
            gdal_installation='gdal_bin')
    basetest_export_by_location(
            input_to_select_from_path, input_to_compare_with_path, output_path, 
            gdal_installation='gdal_default')
        
def test_export_by_location_shp(tmpdir):
    # Prepare input and output paths
    input_to_select_from_path = test_helper.TestFiles.polygons_parcels_shp
    input_to_compare_with_path = test_helper.TestFiles.polygons_zones_gpkg
    output_path = Path(tmpdir) / 'parcels-2020_loc_zones.shp'

    # Try both with and without gdal_bin set
    basetest_export_by_location(
            input_to_select_from_path, input_to_compare_with_path, output_path, 
            gdal_installation='gdal_bin')
    basetest_export_by_location(
            input_to_select_from_path, input_to_compare_with_path, output_path, 
            gdal_installation='gdal_default')

def basetest_export_by_location(
        input_to_select_from_path: Path, 
        input_to_compare_with_path: Path, 
        output_basepath: Path, 
        gdal_installation: str):

    # Do operation
    output_path = output_basepath.parent / f"{output_basepath.stem}_{gdal_installation}{output_basepath.suffix}"
    with test_helper.GdalBin(gdal_installation):
        ok_expected = test_helper.check_runtime_dependencies_ok('twolayer', gdal_installation)
        try:
            geofileops_sql.export_by_location(
                    input_to_select_from_path=input_to_select_from_path,
                    input_to_compare_with_path=input_to_compare_with_path,
                    output_path=output_path)
            test_ok = True
        except MissingRuntimeDependencyError:
            test_ok = False
    assert test_ok is ok_expected, f"Error: for {gdal_installation}, test_ok: {test_ok}, expected: {ok_expected}"

    # If it is expected not to be OK, don't do other checks
    if ok_expected is False:
        return

    # Now check if the tmp file is correctly created
    assert output_path.exists() == True
    layerinfo_orig = geofile.get_layerinfo(input_to_select_from_path)
    layerinfo_output = geofile.get_layerinfo(input_to_select_from_path)
    assert layerinfo_orig.featurecount == layerinfo_output.featurecount
    assert len(layerinfo_orig.columns) == len(layerinfo_output.columns)

    # Check geometry type
    assert layerinfo_output.geometrytype == GeometryType.MULTIPOLYGON 

    # Now check the contents of the result file
    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None

def test_export_by_distance_gpkg(tmpdir):
    # Prepare input and output paths
    input_to_select_from_path = test_helper.TestFiles.polygons_parcels_gpkg
    input_to_compare_with_path = test_helper.TestFiles.polygons_zones_gpkg
    output_path = Path(tmpdir) / 'parcels-2020_distance_zones.gpkg'
    
    # Try both with and without gdal_bin set
    basetest_export_by_location(
            input_to_select_from_path, input_to_compare_with_path, output_path, 
            gdal_installation='gdal_bin')
    basetest_export_by_location(
            input_to_select_from_path, input_to_compare_with_path, output_path, 
            gdal_installation='gdal_default')
    
def test_export_by_distance_shp(tmpdir):
    # Prepare input and output paths
    input_to_select_from_path = test_helper.TestFiles.polygons_parcels_shp
    input_to_compare_with_path = test_helper.TestFiles.polygons_zones_gpkg
    output_path = Path(tmpdir) / 'parcels-2020_distance_zones.shp'

    # Try both with and without gdal_bin set
    basetest_export_by_location(
            input_to_select_from_path, input_to_compare_with_path, output_path, 
            gdal_installation='gdal_bin')
    basetest_export_by_location(
            input_to_select_from_path, input_to_compare_with_path, output_path, 
            gdal_installation='gdal_default')

def basetest_export_by_distance(
        input_to_select_from_path: Path, 
        input_to_compare_with_path: Path, 
        output_basepath: Path,
        gdal_installation: str):

    # Do operation
    output_path = output_basepath.parent / f"{output_basepath.stem}_{gdal_installation}{output_basepath.suffix}"
    with test_helper.GdalBin(gdal_installation):
        ok_expected = test_helper.check_runtime_dependencies_ok('twolayer', gdal_installation)
        try:
            geofileops_sql.export_by_distance(
                    input_to_select_from_path=input_to_select_from_path,
                    input_to_compare_with_path=input_to_compare_with_path,
                    max_distance=10,
                    output_path=output_path)
            test_ok = True
        except MissingRuntimeDependencyError:
            test_ok = False
    assert test_ok is ok_expected, f"Error: for {gdal_installation}, test_ok: {test_ok}, expected: {ok_expected}"

    # If it is expected not to be OK, don't do other checks
    if ok_expected is False:
        return

    # Now check if the tmp file is correctly created
    assert output_path.exists() == True
    layerinfo_orig = geofile.get_layerinfo(input_to_select_from_path)
    layerinfo_output = geofile.get_layerinfo(input_to_select_from_path)
    assert layerinfo_orig.featurecount == layerinfo_output.featurecount
    assert len(layerinfo_orig.columns) == len(layerinfo_output.columns)

    # Check geometry type
    assert layerinfo_output.geometrytype == GeometryType.MULTIPOLYGON 

    # Now check the contents of the result file
    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None

def test_intersect_gpkg(tmpdir):
    # Prepare input and output paths
    input1_path = test_helper.TestFiles.polygons_parcels_gpkg
    input2_path = test_helper.TestFiles.polygons_zones_gpkg
    output_path = Path(tmpdir) / 'parcels-2020_intersect_zones.gpkg'

    # Try both with and without gdal_bin set
    #basetest_intersect(input1_path, input2_path, output_path, gdal_installation='gdal_default')
    basetest_intersect(input1_path, input2_path, output_path, gdal_installation='gdal_bin')
    
def test_intersect_shp(tmpdir):
    # Prepare input and output paths
    input1_path = test_helper.TestFiles.polygons_parcels_shp
    input2_path = test_helper.TestFiles.polygons_zones_gpkg
    output_path = Path(tmpdir) / 'parcels-2020_intersect_zones.gpkg'
    
    # Try both with and without gdal_bin set
    basetest_intersect(input1_path, input2_path, output_path, gdal_installation='gdal_default')
    basetest_intersect(input1_path, input2_path, output_path, gdal_installation='gdal_bin')
    
def basetest_intersect(
        input1_path: Path, 
        input2_path: Path, 
        output_basepath: Path, 
        gdal_installation: str):

    # Do operation
    output_path = output_basepath.parent / f"{output_basepath.stem}_{gdal_installation}{output_basepath.suffix}"
    with test_helper.GdalBin(gdal_installation):
        ok_expected = test_helper.check_runtime_dependencies_ok('twolayer', gdal_installation)
        try:
            geofileops_sql.intersect(
                    input1_path=input1_path,
                    input2_path=input2_path,
                    output_path=output_path,
                    nb_parallel=2)
            test_ok = True
        except MissingRuntimeDependencyError:
            test_ok = False
    assert test_ok is ok_expected, f"Error: for {gdal_installation}, test_ok: {test_ok}, expected: {ok_expected}"

    # If it is expected not to be OK, don't do other checks
    if ok_expected is False:
        return

    # Now check if the tmp file is correctly created
    assert output_path.exists() == True
    layerinfo_input1 = geofile.get_layerinfo(input1_path)
    layerinfo_input2 = geofile.get_layerinfo(input2_path)
    layerinfo_output = geofile.get_layerinfo(output_path)
    assert layerinfo_output.featurecount == 28
    assert (len(layerinfo_input1.columns) + len(layerinfo_input2.columns)) == len(layerinfo_output.columns)

    # Check geometry type
    if output_path.suffix.lower() == '.shp':
        # For shapefiles the type stays POLYGON anyway 
        assert layerinfo_output.geometrytype == GeometryType.POLYGON 
    elif output_path.suffix.lower() == '.gpkg':
        assert layerinfo_output.geometrytype == GeometryType.MULTIPOLYGON 

    # Now check the contents of the result file
    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None

def test_join_by_location_gpkg(tmpdir):
    # Prepare input and output paths
    input1_path = test_helper.TestFiles.polygons_parcels_gpkg
    input2_path = test_helper.TestFiles.polygons_zones_gpkg
    output_path = Path(tmpdir) / 'parcels-2020_join_zones.gpkg'
    
    # Try both with and without gdal_bin set
    basetest_join_by_location(input1_path, input2_path, output_path, gdal_installation='gdal_default')
    basetest_join_by_location(input1_path, input2_path, output_path, gdal_installation='gdal_bin')
    
def test_join_by_location_shp(tmpdir):
    # Prepare input and output paths
    input1_path = test_helper.TestFiles.polygons_parcels_shp
    input2_path = test_helper.TestFiles.polygons_zones_gpkg
    output_path = Path(tmpdir) / 'parcels-2020_join_zones.gpkg'
    
    # Try both with and without gdal_bin set
    basetest_join_by_location(input1_path, input2_path, output_path, gdal_installation='gdal_default')
    basetest_join_by_location(input1_path, input2_path, output_path, gdal_installation='gdal_bin')
    
def basetest_join_by_location(
        input1_path: Path, input2_path: Path,
        output_basepath: Path, 
        gdal_installation: str):
        
    ### Test 1: inner join, intersect
    output_path = output_basepath.parent / f"{output_basepath.stem}_{gdal_installation}_test1_{output_basepath.suffix}"
    with test_helper.GdalBin(gdal_installation):
        ok_expected = test_helper.check_runtime_dependencies_ok('twolayer', gdal_installation)
        try:
            geofileops_sql.join_by_location(
                    input1_path=input1_path,
                    input2_path=input2_path,
                    output_path=output_path,
                    discard_nonmatching=True,
                    force=True)
            test_ok = True
        except MissingRuntimeDependencyError:
            test_ok = False
    assert test_ok is ok_expected, f"Error: for {gdal_installation}, test_ok: {test_ok}, expected: {ok_expected}"

    # If it is expected not to be OK, don't do other checks
    if ok_expected is False:
        return

    # Now check if the output file is correctly created
    assert output_path.exists() == True
    layerinfo_input1 = geofile.get_layerinfo(input1_path)
    layerinfo_input2 = geofile.get_layerinfo(input2_path)
    layerinfo_output = geofile.get_layerinfo(output_path)
    assert layerinfo_output.featurecount == 28
    assert (len(layerinfo_input1.columns) + len(layerinfo_input2.columns) + 1) == len(layerinfo_output.columns)

    # Check geometry type
    if output_path.suffix.lower() == '.shp':
        # For shapefiles the type stays POLYGON anyway 
        assert layerinfo_output.geometrytype == GeometryType.POLYGON 
    elif output_path.suffix.lower() == '.gpkg':
        assert layerinfo_output.geometrytype == GeometryType.MULTIPOLYGON 

    # Now check the contents of the result file
    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None

    ### Test 2: left outer join, intersect
    output_path = output_basepath.parent / f"{output_basepath.stem}_{gdal_installation}_test2_{output_basepath.suffix}"
    with test_helper.GdalBin(gdal_installation):
        ok_expected = test_helper.check_runtime_dependencies_ok('twolayer', gdal_installation)
        try:
            geofileops_sql.join_by_location(
                    input1_path=input1_path,
                    input2_path=input2_path,
                    output_path=output_path,
                    discard_nonmatching=False,
                    force=True)
            test_ok = True
        except MissingRuntimeDependencyError:
            test_ok = False
    assert test_ok is ok_expected, f"Error: for {gdal_installation}, test_ok: {test_ok}, expected: {ok_expected}"

    # If it is expected not to be OK, don't do other checks
    if ok_expected is False:
        return

    # Now check if the output file is correctly created
    assert output_path.exists() == True
    layerinfo_output = geofile.get_layerinfo(output_path)
    assert layerinfo_output.featurecount == 48
    assert (len(layerinfo_input1.columns) + len(layerinfo_input2.columns) + 1) == len(layerinfo_output.columns)

    # Check geometry type
    if output_path.suffix.lower() == '.shp':
        # For shapefiles the type stays POLYGON anyway 
        assert layerinfo_output.geometrytype == GeometryType.POLYGON 
    elif output_path.suffix.lower() == '.gpkg':
        assert layerinfo_output.geometrytype == GeometryType.MULTIPOLYGON

    # Now check the contents of the result file
    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None

def test_join_nearest_gpkg(tmpdir):
    # Prepare input and output paths
    input1_path = test_helper.TestFiles.polygons_parcels_gpkg
    input2_path = test_helper.TestFiles.polygons_zones_gpkg
    output_path = Path(tmpdir) / 'parcels-2020_nearest_zones.gpkg'
    
    # Try both with and without gdal_bin set
    basetest_join_nearest(input1_path, input2_path, output_path, gdal_installation='gdal_default')
    basetest_join_nearest(input1_path, input2_path, output_path, gdal_installation='gdal_bin')
    
def test_join_nearest_shp(tmpdir):
    # Prepare input and output paths
    input1_path = test_helper.TestFiles.polygons_parcels_shp
    input2_path = test_helper.TestFiles.polygons_zones_gpkg
    output_path = Path(tmpdir) / 'parcels-2020_nearest_zones.gpkg'
    
    # Try both with and without gdal_bin set
    basetest_join_nearest(input1_path, input2_path, output_path, gdal_installation='gdal_default')
    basetest_join_nearest(input1_path, input2_path, output_path, gdal_installation='gdal_bin')
    
def basetest_join_nearest(
        input1_path: Path, 
        input2_path: Path,
        output_basepath: Path, 
        gdal_installation: str):
        
    ### Test 1: inner join, intersect
    nb_nearest = 2
    output_path = output_basepath.parent / f"{output_basepath.stem}_{gdal_installation}_test1_{output_basepath.suffix}"
    with test_helper.GdalBin(gdal_installation):
        ok_expected = test_helper.check_runtime_dependencies_ok('twolayer', gdal_installation)
        try:
            geofileops_sql.join_nearest(
                    input1_path=input1_path,
                    input2_path=input2_path,
                    output_path=output_path,
                    nb_nearest=nb_nearest,
                    force=True)
            test_ok = True
        except MissingRuntimeDependencyError:
            test_ok = False
    assert test_ok is ok_expected, f"Error: for {gdal_installation}, test_ok: {test_ok}, expected: {ok_expected}"

    # If it is expected not to be OK, don't do other checks
    if ok_expected is False:
        return

    # Now check if the output file is correctly created
    assert output_path.exists() == True
    layerinfo_input1 = geofile.get_layerinfo(input1_path)
    layerinfo_input2 = geofile.get_layerinfo(input2_path)
    layerinfo_output = geofile.get_layerinfo(output_path)
    assert layerinfo_output.featurecount == nb_nearest * layerinfo_input1.featurecount
    assert (len(layerinfo_input1.columns) + len(layerinfo_input2.columns) + 2) == len(layerinfo_output.columns)

    # Check geometry type
    if output_path.suffix.lower() == '.shp':
        # For shapefiles the type stays POLYGON anyway 
        assert layerinfo_output.geometrytype == GeometryType.POLYGON 
    elif output_path.suffix.lower() == '.gpkg':
        assert layerinfo_output.geometrytype == GeometryType.MULTIPOLYGON 

    # Now check the contents of the result file
    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None

def test_select_two_layers_gpkg(tmpdir):
    # Prepare input and output paths
    input1_path = test_helper.TestFiles.polygons_parcels_gpkg
    input2_path = test_helper.TestFiles.polygons_zones_gpkg
    output_path = Path(tmpdir) / 'parcels-2020_select_zones.gpkg'

    # Try both with and without gdal_bin set
    basetest_select_two_layers(input1_path, input2_path, output_path, gdal_installation='gdal_default')
    basetest_select_two_layers(input1_path, input2_path, output_path, gdal_installation='gdal_bin')
    
def test_select_two_layers_shp(tmpdir):
    # Prepare input and output paths
    input1_path = test_helper.TestFiles.polygons_parcels_shp
    input2_path = test_helper.TestFiles.polygons_zones_gpkg
    output_path = Path(tmpdir) / 'parcels-2020_select_zones.gpkg'
    
    # Try both with and without gdal_bin set
    basetest_select_two_layers(input1_path, input2_path, output_path, gdal_installation='gdal_default')
    basetest_select_two_layers(input1_path, input2_path, output_path, gdal_installation='gdal_bin')
    
def basetest_select_two_layers(
        input1_path: Path, 
        input2_path: Path, 
        output_basepath: Path, 
        gdal_installation: str):

    # Prepare query to execute. At the moment this is just the query for the 
    # intersect() operation.
    input1_layer_info = geofile.get_layerinfo(input1_path)
    input2_layer_info = geofile.get_layerinfo(input2_path)
    primitivetype_to_extract = PrimitiveType(min(
            input1_layer_info.geometrytype.to_primitivetype.value, 
            input2_layer_info.geometrytype.to_primitivetype.value))
    sql_stmt = f'''
            SELECT ST_CollectionExtract(
                    ST_Intersection(layer1.{{input1_geometrycolumn}}, layer2.{{input2_geometrycolumn}}), 
                    {primitivetype_to_extract.value}) as geom
                    {{layer1_columns_prefix_alias_str}}
                    {{layer2_columns_prefix_alias_str}}
                    ,CASE 
                        WHEN layer2.naam = 'zone1' THEN 'in_zone1'
                        ELSE 'niet_in_zone1'
                        END AS category
                FROM {{input1_databasename}}."{{input1_tmp_layer}}" layer1
                JOIN {{input1_databasename}}."rtree_{{input1_tmp_layer}}_{{input1_geometrycolumn}}" layer1tree ON layer1.fid = layer1tree.id
                JOIN {{input2_databasename}}."{{input2_tmp_layer}}" layer2
                JOIN {{input2_databasename}}."rtree_{{input2_tmp_layer}}_{{input2_geometrycolumn}}" layer2tree ON layer2.fid = layer2tree.id
            WHERE 1=1
                {{batch_filter}}
                AND layer1tree.minx <= layer2tree.maxx AND layer1tree.maxx >= layer2tree.minx
                AND layer1tree.miny <= layer2tree.maxy AND layer1tree.maxy >= layer2tree.miny
                AND ST_Intersects(layer1.{{input1_geometrycolumn}}, layer2.{{input2_geometrycolumn}}) = 1
                AND ST_Touches(layer1.{{input1_geometrycolumn}}, layer2.{{input2_geometrycolumn}}) = 0
            '''

    output_path = output_basepath.parent / f"{output_basepath.stem}_{gdal_installation}{output_basepath.suffix}"
    with test_helper.GdalBin(gdal_installation):
        ok_expected = test_helper.check_runtime_dependencies_ok('twolayer', gdal_installation)
        try:
            geofileops_sql.select_two_layers(
                    input1_path=input1_path,
                    input2_path=input2_path,
                    output_path=output_path,
                    sql_stmt=sql_stmt,
                    verbose=True)
            test_ok = True
        except MissingRuntimeDependencyError:
            test_ok = False
    assert test_ok is ok_expected, f"Error: for {gdal_installation}, test_ok: {test_ok}, expected: {ok_expected}"

    # If it is expected not to be OK, don't do other checks
    if ok_expected is False:
        return

    # Now check if the tmp file is correctly created
    assert output_path.exists() == True
    layerinfo_input1 = geofile.get_layerinfo(input1_path)
    layerinfo_input2 = geofile.get_layerinfo(input2_path)
    layerinfo_output = geofile.get_layerinfo(output_path)
    assert layerinfo_output.featurecount == 28
    assert (len(layerinfo_input1.columns) + len(layerinfo_input2.columns) + 1) == len(layerinfo_output.columns)

    # Check geometry type
    if output_path.suffix.lower() == '.shp':
        # For shapefiles the type stays POLYGON anyway 
        assert layerinfo_output.geometrytype == GeometryType.POLYGON 
    elif output_path.suffix.lower() == '.gpkg':
        assert layerinfo_output.geometrytype == GeometryType.MULTIPOLYGON

    # Now check the contents of the result file
    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None

def test_split_gpkg(tmpdir):
    # Prepare input and output paths
    input1_path = test_helper.TestFiles.polygons_parcels_gpkg
    input2_path = test_helper.TestFiles.polygons_zones_gpkg
    output_path = Path(tmpdir) / 'parcels-2020_split_zones.gpkg'

    # Try both with and without gdal_bin set
    basetest_split_layers(input1_path, input2_path, output_path, gdal_installation='gdal_default')
    basetest_split_layers(input1_path, input2_path, output_path, gdal_installation='gdal_bin')
    
def test_split_shp(tmpdir):
    # Prepare input and output paths
    input1_path = test_helper.TestFiles.polygons_parcels_shp
    input2_path = test_helper.TestFiles.polygons_zones_gpkg
    output_path = Path(tmpdir) / 'parcels-2020_split_zones.gpkg'
    
    # Try both with and without gdal_bin set
    basetest_split_layers(input1_path, input2_path, output_path, gdal_installation='gdal_default')
    basetest_split_layers(input1_path, input2_path, output_path, gdal_installation='gdal_bin')
    
def basetest_split_layers(
        input1_path: Path, 
        input2_path: Path, 
        output_basepath: Path, 
        gdal_installation: str):

    # Do operation
    output_path = output_basepath.parent / f"{output_basepath.stem}_{gdal_installation}{output_basepath.suffix}"
    with test_helper.GdalBin(gdal_installation):
        ok_expected = test_helper.check_runtime_dependencies_ok('twolayer', gdal_installation)
        try:
            geofileops_sql.split(
                    input1_path=input1_path,
                    input2_path=input2_path,
                    output_path=output_path,
                    verbose=True)
            test_ok = True
        except MissingRuntimeDependencyError:
            test_ok = False
    assert test_ok is ok_expected, f"Error: for {gdal_installation}, test_ok: {test_ok}, expected: {ok_expected}"

    # If it is expected not to be OK, don't do other checks
    if ok_expected is False:
        return

    # Now check if the tmp file is correctly created
    assert output_path.exists() == True
    layerinfo_input1 = geofile.get_layerinfo(input1_path)
    layerinfo_input2 = geofile.get_layerinfo(input2_path)
    layerinfo_output = geofile.get_layerinfo(output_path)
    assert layerinfo_output.featurecount == 65
    assert (len(layerinfo_input1.columns) + len(layerinfo_input2.columns)) == len(layerinfo_output.columns)

    # Check geometry type
    if output_path.suffix.lower() == '.shp':
        # For shapefiles the type stays POLYGON anyway 
        assert layerinfo_output.geometrytype == GeometryType.POLYGON 
    elif output_path.suffix.lower() == '.gpkg':
        assert layerinfo_output.geometrytype == GeometryType.MULTIPOLYGON

    # Now check the contents of the result file
    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None

def test_union_gpkg(tmpdir):
    ##### Run some tests on parcels versus zones #####
    # Prepare input and output paths
    input1_path = test_helper.TestFiles.polygons_parcels_gpkg
    input2_path = test_helper.TestFiles.polygons_zones_gpkg
    output_path = Path(tmpdir) / 'parcels-2020_union_zones.gpkg'

    # Try both with and without gdal_bin set
    basetest_union(input1_path, input2_path, output_path, gdal_installation='gdal_default')
    basetest_union(input1_path, input2_path, output_path, gdal_installation='gdal_bin')
    
    ##### Also run some tests on basic data with circles #####
    ### Union the single circle towards the 2 circles ###
    input1_path = test_helper.TestFiles.polygons_overlappingcircles_one_gpkg
    input2_path = test_helper.TestFiles.polygons_overlappingcircles_twothree_gpkg
    output_path = Path(tmpdir) / f"{input1_path.stem}_union_{input2_path.stem}.gpkg"
    geofileops_sql.union( 
            input1_path=input1_path,
            input2_path=input2_path,
            output_path=output_path,
            verbose=True)

    # Now check if the tmp file is correctly created
    assert output_path.exists() == True
    layerinfo_input1 = geofile.get_layerinfo(input1_path)
    layerinfo_input2 = geofile.get_layerinfo(input2_path)
    layerinfo_output = geofile.get_layerinfo(output_path)
    assert layerinfo_output.featurecount == 5
    assert (len(layerinfo_input1.columns) + len(layerinfo_input2.columns)) == len(layerinfo_output.columns)

    # Check geometry type
    if output_path.suffix.lower() == '.shp':
        # For shapefiles the type stays POLYGON anyway 
        assert layerinfo_output.geometrytype == GeometryType.POLYGON 
    elif output_path.suffix.lower() == '.gpkg':
        assert layerinfo_output.geometrytype == GeometryType.MULTIPOLYGON

    # Now check the contents of the result file
    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None

    ### Union the two circles towards the single circle ###
    input1_path = test_helper.TestFiles.polygons_overlappingcircles_twothree_gpkg
    input2_path = test_helper.TestFiles.polygons_overlappingcircles_one_gpkg
    output_path = Path(tmpdir) / f"{input1_path.stem}_union_{input2_path.stem}.gpkg"
    geofileops_sql.union( 
            input1_path=input1_path,
            input2_path=input2_path,
            output_path=output_path,
            verbose=True)

    # Now check if the tmp file is correctly created
    assert output_path.exists() == True
    layerinfo_input1 = geofile.get_layerinfo(input1_path)
    layerinfo_input2 = geofile.get_layerinfo(input2_path)
    layerinfo_output = geofile.get_layerinfo(output_path)
    assert layerinfo_output.featurecount == 5
    assert (len(layerinfo_input1.columns) + len(layerinfo_input2.columns)) == len(layerinfo_output.columns)

    # Check geometry type
    if output_path.suffix.lower() == '.shp':
        # For shapefiles the type stays POLYGON anyway 
        assert layerinfo_output.geometrytype == GeometryType.POLYGON 
    elif output_path.suffix.lower() == '.gpkg':
        assert layerinfo_output.geometrytype == GeometryType.MULTIPOLYGON

    # Now check the contents of the result file
    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None

def test_union_shp(tmpdir):
    # Prepare input and output paths
    input1_path = test_helper.TestFiles.polygons_parcels_shp
    input2_path = test_helper.TestFiles.polygons_zones_gpkg
    output_path = Path(tmpdir) / 'parcels-2020_union_zones.gpkg'
    
    # Try both with and without gdal_bin set
    basetest_union(input1_path, input2_path, output_path, gdal_installation='gdal_default')
    basetest_union(input1_path, input2_path, output_path, gdal_installation='gdal_bin')
    
def basetest_union(
        input1_path: Path, 
        input2_path: Path, 
        output_basepath: Path, 
        gdal_installation: str):

    # Do operation
    output_path = output_basepath.parent / f"{output_basepath.stem}_{gdal_installation}{output_basepath.suffix}"
    with test_helper.GdalBin(gdal_installation):
        ok_expected = test_helper.check_runtime_dependencies_ok('twolayer', gdal_installation)
        try:
            geofileops_sql.union(
                    input1_path=input1_path,
                    input2_path=input2_path,
                    output_path=output_path,
                    verbose=True)
            test_ok = True
        except MissingRuntimeDependencyError:
            test_ok = False
    assert test_ok is ok_expected, f"Error: for {gdal_installation}, test_ok: {test_ok}, expected: {ok_expected}"

    # If it is expected not to be OK, don't do other checks
    if ok_expected is False:
        return

    # Now check if the tmp file is correctly created
    assert output_path.exists() == True
    layerinfo_input1 = geofile.get_layerinfo(input1_path)
    layerinfo_input2 = geofile.get_layerinfo(input2_path)
    layerinfo_output = geofile.get_layerinfo(output_path)
    assert layerinfo_output.featurecount == 69
    assert (len(layerinfo_input1.columns) + len(layerinfo_input2.columns)) == len(layerinfo_output.columns)

    # Check geometry type
    if output_path.suffix.lower() == '.shp':
        # For shapefiles the type stays POLYGON anyway 
        assert layerinfo_output.geometrytype == GeometryType.POLYGON 
    elif output_path.suffix.lower() == '.gpkg':
        assert layerinfo_output.geometrytype == GeometryType.MULTIPOLYGON

    # Now check the contents of the result file
    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None

if __name__ == '__main__':
    # Init
    tmpdir = test_helper.init_test_for_debug(Path(__file__).stem)

    # Two layer operations
    #test_erase_gpkg(tmpdir)
    #test_erase_shp(tmpdir)
    #test_intersect_gpkg(tmpdir)
    #test_export_by_distance_shp(tmpdir)
    #test_export_by_location_gpkg(tmpdir)
    #test_join_by_location_gpkg(tmpdir)
    #test_select_two_layers_gpkg(tmpdir)
    #test_split_gpkg(tmpdir)
    test_union_gpkg(tmpdir)
    


# -*- coding: utf-8 -*-
"""
Tests for operations that are executed using a sql statement on one layer.
"""

from pathlib import Path
import sys

# Add path so the local geofileops packages are found 
sys.path.insert(0, str(Path(__file__).resolve().parent.parent))
from geofileops import geofile
from geofileops.geofile import GeometryType
from geofileops.util.general_util import MissingRuntimeDependencyError 
from geofileops.util import geofileops_sql
from geofileops.util import ogr_util
from tests import test_helper

def test_buffer_gpkg(tmpdir):
    # Buffer to test dir, and try with and without gdal_bin set
    input_path = test_helper.TestFiles.polygons_parcels_gpkg
    output_path = Path(tmpdir) / 'polygons_parcels-output.gpkg'
    basetest_buffer(input_path, output_path, gdal_installation='gdal_default')
    basetest_buffer(input_path, output_path, gdal_installation='gdal_bin')

    # Buffer point source to test dir
    input_path = test_helper.TestFiles.points_gpkg
    output_path = Path(tmpdir) / 'points-output.gpkg'
    basetest_buffer(input_path, output_path, gdal_installation='gdal_default')
    basetest_buffer(input_path, output_path, gdal_installation='gdal_bin')

    # Buffer line source to test dir
    input_path = test_helper.TestFiles.linestrings_rows_of_trees_gpkg
    output_path = Path(tmpdir) / 'linestrings_rows_of_trees-output.gpkg'
    basetest_buffer(input_path, output_path, gdal_installation='gdal_default')
    basetest_buffer(input_path, output_path, gdal_installation='gdal_bin')

def test_buffer_shp(tmpdir):
    # Buffer to test dir
    input_path = test_helper.TestFiles.polygons_parcels_shp
    output_path = Path(tmpdir) / 'polygons_parcels-output.shp'

    # Try both with and without gdal_bin set
    basetest_buffer(input_path, output_path, gdal_installation='gdal_default')
    basetest_buffer(input_path, output_path, gdal_installation='gdal_bin')
    
def basetest_buffer(
        input_path: Path, 
        output_basepath: Path, 
        gdal_installation: str):
    
    # Do operation
    output_path = output_basepath.parent / f"{output_basepath.stem}_{gdal_installation}{output_basepath.suffix}"
    with test_helper.GdalBin(gdal_installation):
        ok_expected = test_helper.check_runtime_dependencies_ok('buffer', gdal_installation)
        try:
            geofileops_sql.buffer(input_path=input_path, output_path=output_path, distance=1)
            test_ok = True
        except MissingRuntimeDependencyError:
            test_ok = False
    assert test_ok is ok_expected, f"Error: for {gdal_installation}, test_ok: {test_ok}, expected: {ok_expected}"

    # If it is expected not to be OK, don't do other checks
    if ok_expected is False:
        return

    # Now check if the tmp file is correctly created
    layerinfo_orig = geofile.get_layerinfo(input_path)
    layerinfo_output = geofile.get_layerinfo(output_path)
    assert layerinfo_orig.featurecount == layerinfo_output.featurecount
    assert len(layerinfo_orig.columns) == len(layerinfo_output.columns)

    # Check geometry type
    assert layerinfo_output.geometrytype == GeometryType.MULTIPOLYGON 

    # Now check the contents of the result file
    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None
    geofile.remove(output_path)

def test_convexhull_gpkg(tmpdir):
    # Execute to test dir
    input_path = test_helper.TestFiles.polygons_parcels_gpkg
    output_path = Path(tmpdir) / 'polygons_parcels-output.gpkg'
    
    # Try both with and without gdal_bin set
    basetest_convexhull(input_path, output_path, gdal_installation='gdal_bin')
    basetest_convexhull(input_path, output_path, gdal_installation='gdal_default')

def test_convexhull_shp(tmpdir):
    # Select some data from src to tmp file
    input_path = test_helper.TestFiles.polygons_parcels_shp
    output_path = Path(tmpdir) / 'polygons_parcels-output.shp'

    # Try both with and without gdal_bin set
    basetest_convexhull(input_path, output_path, gdal_installation='gdal_bin')
    basetest_convexhull(input_path, output_path, gdal_installation='gdal_default')

def basetest_convexhull(
        input_path: Path, 
        output_basepath: Path, 
        gdal_installation: str):
    
    # Do operation  
    output_path = output_basepath.parent / f"{output_basepath.stem}_{gdal_installation}{output_basepath.suffix}"
    with test_helper.GdalBin(gdal_installation):
        ok_expected = test_helper.check_runtime_dependencies_ok('', gdal_installation)
        try:
            geofileops_sql.convexhull(input_path=input_path, output_path=output_path)
            test_ok = True
        except MissingRuntimeDependencyError:
            test_ok = False
    assert test_ok is ok_expected, f"Error: for {gdal_installation}, test_ok: {test_ok}, expected: {ok_expected}"

    # If it is expected not to be OK, don't do other checks
    if ok_expected is False:
        return

    # Now check if the tmp file is correctly created
    assert output_path.exists() == True
    layerinfo_orig = geofile.get_layerinfo(input_path)
    layerinfo_output = geofile.get_layerinfo(output_path)
    assert layerinfo_orig.featurecount == layerinfo_output.featurecount
    assert len(layerinfo_orig.columns) == len(layerinfo_output.columns)

    # Check geometry type
    assert layerinfo_output.geometrytype == GeometryType.MULTIPOLYGON 

    # Now check the contents of the result file
    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None

def test_isvalid_gpkg(tmpdir):
    # Buffer to test dir
    input_path = test_helper.TestFiles.polygons_parcels_gpkg
    output_path = Path(tmpdir) / 'polygons_parcels-output.gpkg'

    # Try both with and without gdal_bin set
    basetest_isvalid(input_path, output_path, gdal_installation='gdal_bin')
    basetest_isvalid(input_path, output_path, gdal_installation='gdal_default')

def test_isvalid_shp(tmpdir):
    # Select some data from src to tmp file
    input_path = test_helper.TestFiles.polygons_parcels_shp
    output_path = Path(tmpdir) / 'polygons_parcels-output.gpkg'
    
    # Try both with and without gdal_bin set
    basetest_isvalid(input_path, output_path, gdal_installation='gdal_bin')
    basetest_isvalid(input_path, output_path, gdal_installation='gdal_default')
    
def basetest_isvalid(
        input_path: Path, 
        output_basepath: Path, 
        gdal_installation: str,
        ok_expected: bool = None):
    
    # Do operation
    output_path = output_basepath.parent / f"{output_basepath.stem}_{gdal_installation}{output_basepath.suffix}"
    with test_helper.GdalBin(gdal_installation):
        if ok_expected is None:
            ok_expected = test_helper.check_runtime_dependencies_ok('isvalid', gdal_installation)
        try:
            geofileops_sql.isvalid(input_path=input_path, output_path=output_path, nb_parallel=2)
            test_ok = True
        except MissingRuntimeDependencyError:
            test_ok = False
    assert test_ok is ok_expected, f"Error: for {gdal_installation}, test_ok: {test_ok}, expected: {ok_expected}"

    # If it is expected not to be OK, don't do other checks
    if ok_expected is False:
        return

    '''
    # Now check if the tmp file is correctly created
    assert output_path.exists() == True
    layerinfo_select = geofile.getlayerinfo(output_path)
    assert layerinfo_orig.featurecount == layerinfo_select.featurecount
    assert (len(layerinfo_orig.columns)+3) == len(layerinfo_select.columns)

    output_gdf = geofile.read_file(output_path)
    print(output_gdf)
    assert output_gdf['geom'][0] is None
    assert output_gdf['isvalid'][0] == 1
    assert output_gdf['isvalidreason'][0] == 'Valid Geometry'
    '''

def test_makevalid_gpkg(tmpdir):
    # makevalid to test dir
    input_path = test_helper.TestFiles.polygons_invalid_geometries_gpkg
    output_path = Path(tmpdir) / f"{input_path.stem}_valid-output.gpkg"
    
    # Try both with and without gdal_bin set
    basetest_makevalid(input_path, output_path, gdal_installation='gdal_bin')
    basetest_makevalid(input_path, output_path, gdal_installation='gdal_default')

def test_makevalid_shp(tmpdir):
    # makevalid to test dir
    input_path = test_helper.TestFiles.polygons_invalid_geometries_shp
    output_path = Path(tmpdir) / f"{input_path.stem}_valid-output.shp"
    
    # Try both with and without gdal_bin set
    basetest_makevalid(input_path, output_path, gdal_installation='gdal_bin')
    basetest_makevalid(input_path, output_path, gdal_installation='gdal_default')
           
def basetest_makevalid(
        input_path: Path, 
        output_basepath: Path, 
        gdal_installation: str,
        ok_expected: bool = None):

    # Do operation
    output_path = output_basepath.parent / f"{output_basepath.stem}_{gdal_installation}{output_basepath.suffix}"
    with test_helper.GdalBin(gdal_installation):
        if ok_expected is None:
            ok_expected = test_helper.check_runtime_dependencies_ok('makevalid', gdal_installation)
        try: 
            geofileops_sql.makevalid(input_path=input_path, output_path=output_path, nb_parallel=2)
            test_ok = True
        except MissingRuntimeDependencyError:
            test_ok = False
        assert test_ok is ok_expected, f"Error: for {gdal_installation}, test_ok: {test_ok}, expected: {ok_expected}"

        # If it is expected not to be OK, don't do other checks
        if ok_expected is False:
            return

        # Now check if the output file is correctly created
        assert output_path.exists() == True
        layerinfo_orig = geofile.get_layerinfo(input_path)
        layerinfo_output = geofile.get_layerinfo(output_path)
        assert layerinfo_orig.featurecount == layerinfo_output.featurecount
        assert len(layerinfo_orig.columns) == len(layerinfo_output.columns)
    
        # Check geometry type
        assert layerinfo_output.geometrytype == GeometryType.MULTIPOLYGON 

        # Now check the contents of the result file
        output_gdf = geofile.read_file(output_path)
        assert output_gdf['geometry'][0] is not None

        # Make sure the input file was not valid
        output_isvalid_path = output_path.parent / f"{output_path.stem}_is-valid{output_path.suffix}"
        isvalid = geofileops_sql.isvalid(input_path=input_path, output_path=output_isvalid_path)
        assert isvalid is False, "Input file should contain invalid features"

        # Check if the result file is valid
        output_new_isvalid_path = output_path.parent / f"{output_path.stem}_new_is-valid{output_path.suffix}"
        isvalid = geofileops_sql.isvalid(input_path=output_path, output_path=output_new_isvalid_path)
        assert isvalid == True, "Output file shouldn't contain invalid features"

def test_select_gpkg(tmpdir):
    # Select some data from src to tmp file
    input_path = test_helper.TestFiles.polygons_parcels_gpkg
    output_path = Path(tmpdir) / 'polygons_parcels-output.gpkg'

    basetest_select(input_path, output_path)

def test_select_gpkg_to_shp(tmpdir):
    # Select some data from src to tmp file
    input_path = test_helper.TestFiles.polygons_parcels_gpkg
    output_path = Path(tmpdir) / 'polygons_parcels-output.shp'

    basetest_select(input_path, output_path)

def test_select_shp(tmpdir):
    # Select some data from src to tmp file
    input_path = test_helper.TestFiles.polygons_parcels_shp
    output_path = Path(tmpdir) / 'polygons_parcels-output.shp'

    basetest_select(input_path, output_path)

def test_select_shp_to_gpkg(tmpdir):
    # Select some data from src to tmp file
    input_path = test_helper.TestFiles.polygons_parcels_shp
    output_path = Path(tmpdir) / 'polygons_parcels-output.gpkg'

    basetest_select(input_path, output_path)

def basetest_select(
        input_path: Path, 
        output_path: Path):

    layerinfo_input = geofile.get_layerinfo(input_path)
    sql_stmt = 'SELECT {geometrycolumn}, oidn, uidn FROM "{input_layer}"'
    geofileops_sql.select(
            input_path=input_path,
            output_path=output_path,
            sql_stmt=sql_stmt)

    # Now check if the tmp file is correctly created
    layerinfo_output = geofile.get_layerinfo(output_path)
    assert layerinfo_input.featurecount == layerinfo_output.featurecount
    assert 'OIDN' in layerinfo_output.columns
    assert 'UIDN' in layerinfo_output.columns
    assert len(layerinfo_output.columns) == 2

    # Check geometry type
    assert layerinfo_output.geometrytype == GeometryType.MULTIPOLYGON 

    # Now check the contents of the result file
    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None

def test_select_various_options_shp(tmpdir):
    # Select some data from src to tmp file
    input_path = test_helper.TestFiles.polygons_parcels_shp
    output_path = Path(tmpdir) / 'polygons_parcels-output.shp'

    basetest_select_various_options(input_path, output_path)

def test_select_various_options_gpkg(tmpdir):
    # Select some data from src to tmp file
    input_path = test_helper.TestFiles.polygons_parcels_gpkg
    output_path = Path(tmpdir) / 'polygons_parcels-output.gpkg'

    basetest_select_various_options(input_path, output_path)
    
def basetest_select_various_options(
        input_path: Path, 
        output_path: Path):

    ### Check if columns parameter works (case insensitive) ###
    columns = ['OIDN', 'uidn', 'HFDTLT', 'lblhfdtlt', 'GEWASGROEP', 'lengte', 'OPPERVL']
    layerinfo_input = geofile.get_layerinfo(input_path)
    sql_stmt = '''SELECT {geometrycolumn}
                        {columns_to_select_str} 
                    FROM "{input_layer}"'''
    geofileops_sql.select(
            input_path=input_path,
            output_path=output_path,
            columns=columns,
            sql_stmt=sql_stmt)

    # Now check if the tmp file is correctly created
    layerinfo_select = geofile.get_layerinfo(output_path)
    assert layerinfo_input.featurecount == layerinfo_select.featurecount
    assert 'OIDN' in layerinfo_select.columns
    assert 'UIDN' in layerinfo_select.columns
    assert len(layerinfo_select.columns) == len(columns)

    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None

    ### Check if ... parameter works ###
    # TODO: increase test coverage of other options...

def test_simplify_gpkg(tmpdir):
    # Simplify polygon source to test dir, with and without gdal_bin set
    input_path = test_helper.TestFiles.polygons_parcels_gpkg
    output_path = Path(tmpdir) / input_path.name
    basetest_simplify(input_path, output_path, 
            expected_output_geometrytype=GeometryType.MULTIPOLYGON,
            gdal_installation='gdal_default')
    basetest_simplify(input_path, output_path, 
            expected_output_geometrytype=GeometryType.MULTIPOLYGON,
            gdal_installation='gdal_bin')

    # Simplify point source to test dir
    input_path = test_helper.TestFiles.points_gpkg
    output_path = Path(tmpdir) / 'points-output.gpkg'
    basetest_simplify(input_path, output_path, 
            expected_output_geometrytype=GeometryType.MULTIPOINT,
            gdal_installation='gdal_default')
    basetest_simplify(input_path, output_path, 
            expected_output_geometrytype=GeometryType.MULTIPOINT,
            gdal_installation='gdal_bin')

    # Simplify line source to test dir
    input_path = test_helper.TestFiles.linestrings_rows_of_trees_gpkg
    output_path = Path(tmpdir) / 'linestrings_rows_of_trees-output.gpkg'
    basetest_simplify(input_path, output_path, 
            expected_output_geometrytype=GeometryType.MULTILINESTRING,
            gdal_installation='gdal_default')
    basetest_simplify(input_path, output_path, 
            expected_output_geometrytype=GeometryType.MULTILINESTRING,
            gdal_installation='gdal_bin')

def test_simplify_shp(tmpdir):
    # Simplify to test dir
    input_path = test_helper.TestFiles.polygons_parcels_shp
    output_path = Path(tmpdir) / 'polygons_parcels-output.shp'

    # Try both with and without gdal_bin set
    basetest_simplify(input_path, output_path, 
            expected_output_geometrytype=GeometryType.MULTIPOLYGON,
            gdal_installation='gdal_default')
    basetest_simplify(input_path, output_path, 
            expected_output_geometrytype=GeometryType.MULTIPOLYGON,
            gdal_installation='gdal_bin')
    
def basetest_simplify(
        input_path: Path, 
        output_basepath: Path,
        expected_output_geometrytype: GeometryType,
        gdal_installation: str):

    # Do operation
    output_path = output_basepath.parent / f"{output_basepath.stem}_{gdal_installation}{output_basepath.suffix}"
    with test_helper.GdalBin(gdal_installation):
        ok_expected = test_helper.check_runtime_dependencies_ok('', gdal_installation)
        try:
            geofileops_sql.simplify(
                    input_path=input_path, output_path=output_path,
                    tolerance=5)
            test_ok = True
        except MissingRuntimeDependencyError:
            test_ok = False
    assert test_ok is ok_expected, f"Error: for {gdal_installation}, test_ok: {test_ok}, expected: {ok_expected}"

    # If it is expected not to be OK, don't do other checks
    if ok_expected is False:
        return

    # Now check if the output file is correctly created
    assert output_path.exists() == True
    layerinfo_orig = geofile.get_layerinfo(input_path)
    layerinfo_output = geofile.get_layerinfo(output_path)
    assert layerinfo_orig.featurecount == layerinfo_output.featurecount
    assert len(layerinfo_orig.columns) == len(layerinfo_output.columns)

    # Check geometry type
    assert layerinfo_output.geometrytype == expected_output_geometrytype 

    # Now check the contents of the result file
    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None

if __name__ == '__main__':
    # Init
    tmpdir = test_helper.init_test_for_debug(Path(__file__).stem)

    # Single layer operations
    test_buffer_gpkg(tmpdir)
    #test_makevalid_shp(tmpdir)
    #test_makevalid_gpkg(tmpdir)
    #test_isvalid_shp(tmpdir)
    #test_isvalid_gpkg(tmpdir)
    #test_convexhull_gpkg(tmpdir)
    #test_convexhull_shp(tmpdir)
    #test_select_geos_version(tmpdir)
    #test_simplify_gpkg(tmpdir)
    


# -*- coding: utf-8 -*-
"""
Module containing utilities regarding low level vector operations.
"""

import enum
import logging
import math
from typing import Any, List, Optional, Union

import geopandas as gpd
import numpy as np
import pygeos
from shapely.geometry.polygon import Polygon
import shapely.wkb as sh_wkb
import shapely.geometry as sh_geom
import shapely.ops as sh_ops
# Import simplification if it is available. Only throw exception in runtime...
try:
    import simplification.cutil as simplification
except ImportError as ex:
    None

#-------------------------------------------------------------
# First define/init some general variables/constants
#-------------------------------------------------------------
# Get a logger...
logger = logging.getLogger(__name__)
#logger.setLevel(logging.DEBUG)

#-------------------------------------------------------------
# Geometry helpers
#-------------------------------------------------------------

class GeometryType(enum.Enum):
    MISSING = -1
    GEOMETRY = 0
    POINT = 1
    LINESTRING = 2
    POLYGON = 3
    MULTIPOINT = 4
    MULTILINESTRING = 5
    MULTIPOLYGON = 6
    GEOMETRYCOLLECTION = 7

    @classmethod
    def _missing_(cls, value: Union[str, int]):
        """
        Expand options in the Geometrytype() constructor.
        
        Args:
            value (Union[str, int, GeometryType]): 
                * string: lookup using case insensitive name
                * GeometryType: create the same GeometryType as the one passed in

        Returns:
            [GeometryType]: The corresponding GeometryType. 
        """
        if value is None:
            return None
        elif isinstance(value, str):
            # If a string is passed in, try lookup based on case insensitive 
            # enum name
            return cls(GeometryType[value.upper()])
        elif isinstance(value, GeometryType):
            # If a Geometry type is passed in, return same GeometryType
            # TODO: why create a new one?
            return cls(value.value)
        # Default behaviour (= lookup based on value)
        return super()._missing_(value)

    @property
    def is_multitype(self):
        return self in (
                GeometryType.MULTIPOINT, GeometryType.MULTILINESTRING, 
                GeometryType.MULTIPOLYGON, GeometryType.GEOMETRYCOLLECTION)

    @property
    def to_multitype(self):
        if self in [GeometryType.MULTIPOINT, GeometryType.MULTILINESTRING, 
            GeometryType.MULTIPOLYGON, GeometryType.GEOMETRYCOLLECTION]:
            return self
        elif self is GeometryType.POINT:
            return GeometryType.MULTIPOINT
        elif self is GeometryType.LINESTRING:
            return GeometryType.MULTILINESTRING
        elif self is GeometryType.POLYGON:
            return GeometryType.MULTIPOLYGON
        else:
            raise Exception(f"No multitype implemented for: {self}")

    @property
    def to_primitivetype(self):
        # Lookup correct primitive type
        if self in [GeometryType.POINT, GeometryType.MULTIPOINT]:
            return PrimitiveType.POINT
        elif self in [GeometryType.LINESTRING, GeometryType.MULTILINESTRING]:
            return PrimitiveType.LINESTRING
        elif self in [GeometryType.POLYGON, GeometryType.MULTIPOLYGON]:
            return PrimitiveType.POLYGON
        elif self is GeometryType.GEOMETRYCOLLECTION:
            return self
        else:
            raise Exception(f"No primitive type implemented for {self}")
    
class PrimitiveType(enum.Enum):
    POINT = 1
    LINESTRING = 2
    POLYGON = 3

    @classmethod
    def _missing_(cls, value):
        if value is None:
            return None
        elif isinstance(value, str):
            return cls(PrimitiveType[value.upper()])
        elif isinstance(value, PrimitiveType):
            return cls(value.value)
        return super()._missing_(value)

    @property
    def to_multitype(self) -> GeometryType:
        if self is PrimitiveType.POINT:
            return GeometryType.MULTIPOINT
        elif self is PrimitiveType.LINESTRING:
            return GeometryType.MULTILINESTRING
        elif self is PrimitiveType.POLYGON:
            return GeometryType.MULTIPOLYGON
        else:
            raise Exception(f"No multitype implemented for: {self}")

def collection_extract(
        geometry: Optional[sh_geom.base.BaseGeometry],
        primitivetype: PrimitiveType) -> Optional[sh_geom.base.BaseGeometry]:
    """
    Extracts the geometries from the input geom that comply with the 
    primitive_type specified and returns them as (Multi)geometry.

    Args:
        geometry (sh_geom.base.BaseGeometry): geometry to extract the polygons 
            from.
        primitivetype (GeometryPrimitiveTypes): the primitive type to extract
            from the input geom.

    Raises:
        Exception: if in_geom is an unsupported geometry type or the primitive 
            type is invalid.  

    Returns:
        sh_geom.base.BaseGeometry: List of primitive geometries, only 
            containing the primitive type specified.
    """
    # Extract the polygons from the multipolygon, but store them as multipolygons anyway
    if geometry is None:
        return None
    elif isinstance(geometry, sh_geom.Point) or isinstance(geometry, sh_geom.MultiPoint):
        if primitivetype == PrimitiveType.POINT:
            return geometry
    elif isinstance(geometry, sh_geom.LineString) or isinstance(geometry, sh_geom.MultiLineString):
        if primitivetype == PrimitiveType.LINESTRING:
            return geometry
    elif isinstance(geometry, sh_geom.Polygon) or isinstance(geometry, sh_geom.MultiPolygon):
        if primitivetype == PrimitiveType.POLYGON:
            return geometry
    elif isinstance(geometry, sh_geom.GeometryCollection):
        returngeoms = []
        for geometry in list(sh_geom.GeometryCollection(geometry)):
            returngeoms.append(collection_extract(geometry, primitivetype=primitivetype))
        if len(returngeoms) > 0:
            return collect(returngeoms)
    else:
        raise Exception(f"Invalid/unsupported geometry(type): {geometry}")

    # Nothing found yet, so return None
    return None

def collect(
        geometry_list: List[sh_geom.base.BaseGeometry]) -> Optional[sh_geom.base.BaseGeometry]:
    """
    Collect a list of geometries to one geometry. 
    
    Examples:
      * if the list contains only Polygon's, returns a MultiPolygon. 
      * if the list contains different types, returns a GeometryCollection.

    Args:
        geometry_list (List[sh_geom.base.BaseGeometry]): [description]

    Raises:
        Exception: raises an exception if one of the input geometries is of an 
            unknown type. 

    Returns:
        sh_geom.base.BaseGeometry: the result
    """
    # First remove all None geometries in the input list
    geometry_list = [geometry for geometry in geometry_list 
                     if geometry is not None and geometry.is_empty is False]

    # If the list is empty or contains only 1 element, it is easy...
    if geometry_list is None or len(geometry_list) == 0:
        return None
    elif len(geometry_list) == 1:
        return geometry_list[0]
    
    # Loop over all elements in the list, and determine the appropriate geometry type to create
    result_collection_type = GeometryType(geometry_list[0].geom_type).to_multitype
    for geom in geometry_list:
        # If it is the same as the collection_geom_type, continue checking
        if GeometryType(geom.geom_type).to_multitype == result_collection_type:
            continue
        else:
            # If multiple types in the list, result becomes a geometrycollection
            result_collection_type = GeometryType.GEOMETRYCOLLECTION
            break
    
    # Now we can create the collection
    # Explode the multi-geometries to single ones
    singular_geometry_list = []
    for geom in geometry_list:
        if isinstance(geom, sh_geom.base.BaseMultipartGeometry):
            singular_geometry_list.extend(list(geom))
        else:
            singular_geometry_list.append(geom)

    if result_collection_type == GeometryType.MULTIPOINT:
        return sh_geom.MultiPoint(singular_geometry_list)
    elif result_collection_type == GeometryType.MULTILINESTRING:
        return sh_geom.MultiLineString(singular_geometry_list)
    elif result_collection_type == GeometryType.MULTIPOLYGON:
        return sh_geom.MultiPolygon(singular_geometry_list)
    elif result_collection_type == GeometryType.GEOMETRYCOLLECTION:
        return sh_geom.GeometryCollection(geometry_list)
    else:
        raise Exception(f"Unsupported geometry type: {result_collection_type}")

'''
def force_geometrytype(
        geometry: sh_geom.base.BaseGeometry,
        dest_geometrytype: GeometryType) -> sh_geom.base.BaseGeometry:
    # Cast to destination geometrytype
    if dest_geometrytype is GeometryType.MULTIPOLYGON:
        gdf.geometry = [sh_geom.MultiPolygon([feature]) 
                        if type(feature) == sh_geom.Polygon 
                        else feature for feature in gdf.geometry]
    elif dest_geometrytype is GeometryType.MULTIPOINT:
        gdf.geometry = [sh_geom.MultiPoint([feature]) 
                        if type(feature) == sh_geom.Point 
                        else feature for feature in gdf.geometry]
    elif dest_geometrytype is GeometryType.MULTILINESTRING:
        gdf.geometry = [sh_geom.MultiLineString([feature]) 
                        if type(feature) == sh_geom.LineString 
                        else feature for feature in gdf.geometry]
    elif dest_geometrytype in [
            GeometryType.POLYGON, GeometryType.POINT, GeometryType.LINESTRING]:
        logger.debug(f"geometrytype is {dest_geometrytype}, so no conversion is done")
    else:
        raise Exception(f"Unsupported geometrytype: {dest_geometrytype}")
'''

def make_valid(geometry: Optional[sh_geom.base.BaseGeometry]) -> Optional[sh_geom.base.BaseGeometry]:
    """
    Make a geometry valid.

    Args:
        geometry (Optional[sh_geom.base.BaseGeometry]): A (possibly) invalid geometry.

    Returns:
        Optional[sh_geom.base.BaseGeometry]: The fixed geometry.
    """
    if geometry is None:
        return None
    else:
        return sh_wkb.loads(pygeos.io.to_wkb(pygeos.make_valid(pygeos.io.from_shapely(geometry))))

def numberpoints(geometry: Optional[sh_geom.base.BaseGeometry]) -> int:
    """
    Calculates the total number of points in a geometry.

    Args:
        geometry (sh_geom.base.BaseGeometry): the geometry to count the point of.

    Returns:
        int: the number of points in the geometry.
    """
    # If it is a multi-part, recursively call numberpoints for all parts. 
    if geometry is None:
        return 0
    elif isinstance(geometry, sh_geom.base.BaseMultipartGeometry):
        nb_points = 0
        for geom in list(geometry):
            nb_points += numberpoints(geom)
        return nb_points
    elif isinstance(geometry, sh_geom.Polygon):
        # If it is a polygon, calculate number for exterior and interior rings. 
        nb_points = len(geometry.exterior.coords)
        for ring in geometry.interiors:
            nb_points += len(ring.coords)
        return nb_points
    else:
        # For other types, it is just the number of coordinates.
        return len(geometry.coords)

def remove_inner_rings(
        geometry: Union[sh_geom.Polygon, sh_geom.MultiPolygon, None],
        min_area_to_keep: float = None) -> Union[sh_geom.Polygon, sh_geom.MultiPolygon, None]:
    
    # If input geom is None, just return.
    if geometry is None:
        return None
    
    # Define function to treat simple polygons
    def remove_inner_rings_polygon(
            geom_poly: sh_geom.Polygon,
            min_area_to_keep: float = None) -> sh_geom.Polygon:
        # If all inner rings need to be removed...
        if min_area_to_keep is None or min_area_to_keep == 0.0:
            # If there are no interior rings anyway, just return input
            if len(geom_poly.interiors) == 0:
                return geom_poly
            else:
                # Else create new polygon with only the exterior ring
                return sh_ops.Polygon(geom_poly.exterior)
        
        # If only small rings need to be removed... loop over them
        ring_coords_to_keep = []
        small_ring_found = False
        for ring in geom_poly.interiors:
            if abs(sh_ops.Polygon(ring).area) <= min_area_to_keep:
                small_ring_found = True
            else:
                ring_coords_to_keep.append(ring.coords)
        
        # If no small rings were found, just return input
        if small_ring_found == False:
            return geom_poly
        else:
            return sh_ops.Polygon(geom_poly.exterior.coords, ring_coords_to_keep)
    
    # If the input is a simple Polygon, apply remove on it and return.
    if isinstance(geometry, sh_geom.Polygon):
        return remove_inner_rings_polygon(geometry, min_area_to_keep)
    elif isinstance(geometry, sh_geom.MultiPolygon):
        # If the input is a MultiPolygon, apply remove on each Polygon in it. 
        polys = []
        for poly in list(geometry):
            polys.append(remove_inner_rings_polygon(poly, min_area_to_keep))
        return sh_geom.MultiPolygon(polys)
    else:
        raise Exception(f"remove_inner_rings is not possible with geometrytype: {geometry.type}, geometry: {geometry}")

class SimplifyAlgorithm(enum.Enum):
    RAMER_DOUGLAS_PEUCKER = 'rdp'
    LANG = 'lang'
    VISVALINGAM_WHYATT = 'vw'

def simplify_ext(
        geometry: Optional[sh_geom.base.BaseGeometry],
        tolerance: float,
        algorithm: SimplifyAlgorithm = SimplifyAlgorithm.RAMER_DOUGLAS_PEUCKER,
        lookahead: int = 8,
        preserve_topology: bool = True,
        keep_points_on: sh_geom.base.BaseGeometry = None) -> Optional[sh_geom.base.BaseGeometry]:
    """
    Simplify the geometry, with extended options.

    Args:
        geometry (shapely geometry): the geometry to simplify
        tolerance (float): mandatory for the following algorithms:  
            * RAMER_DOUGLAS_PEUCKER: distance to use as tolerance  
            * LANG: distance to use as tolerance
            * VISVALINGAM_WHYATT: area to use as tolerance
        algorithm (SimplifyAlgorithm, optional): algorithm to use.
            Defaults to SimplifyAlgorithm.RAMER_DOUGLAS_PEUCKER.
        lookahead (int, optional): the number of points to consider for removing
            in a moving window. Used for LANG algorithm. Defaults to 8.
        preserve_topology (bool, optional): True to (try to) return valid 
            geometries as result. Defaults to True.
        keep_points_on (BaseGeometry], optional): point of the geometry to 
            that intersect with these geometries are not removed. Defaults to None.

    Raises:
        Exception: [description]
        Exception: [description]
        Exception: [description]

    Returns:
        sh_geom.base.BaseGeometry: The simplified version of the geometry.
    """
    # Init:
    if geometry is None:
        return None
    if algorithm in [SimplifyAlgorithm.RAMER_DOUGLAS_PEUCKER,
                     SimplifyAlgorithm.VISVALINGAM_WHYATT]:
        try:
            import simplification.cutil as simplification
        except ImportError as ex:
            raise ImportError(f"To use simplify_ext using rdp or vw, first do: 'pip install simplification'") from ex

    # Define some inline funtions 
    # Apply the simplification (can result in multipolygons)
    def simplify_polygon(polygon: sh_geom.Polygon) -> Union[sh_geom.Polygon, sh_geom.MultiPolygon, None]:
        ### First simplify exterior ring ###
        exterior_simplified = simplify_coords(polygon.exterior.coords)
        
        # If topology needs to be preserved, keep original ring if simplify results in 
        # None or not enough points   
        if(preserve_topology is True
           and(exterior_simplified is None 
               or len(exterior_simplified) < 3)):
            exterior_simplified = polygon.exterior.coords

        ### Now simplify interior rings ###
        interiors_simplified = []
        for interior in polygon.interiors:
            interior_simplified = simplify_coords(interior.coords)
            
            # If simplified version is ring, add it
            if(interior_simplified is not None 
               and len(interior_simplified) >= 3):
                interiors_simplified.append(interior_simplified) 
            elif preserve_topology is True:
                # If result is no ring, but topology needs to be preserved, add original ring
                interiors_simplified.append(interior.coords)              

        result_poly = sh_geom.Polygon(exterior_simplified, interiors_simplified)
        
        # Extract only polygons as result + try to make valid
        result_poly = collection_extract(make_valid(result_poly), primitivetype=PrimitiveType.POLYGON)
        
        # If the result is None and the topology needs to be preserved, return original polygon
        if preserve_topology is True and result_poly is None: 
            return polygon
        
        # evade pyLance warning + return
        assert(result_poly is None
               or isinstance(result_poly, sh_geom.Polygon)
               or isinstance(result_poly, sh_geom.MultiPolygon))
        return result_poly

    def simplify_linestring(linestring: sh_geom.LineString) -> sh_geom.LineString:
        # If the linestring cannot be simplified, return it
        if linestring is None or len(linestring.coords) <= 2:
            return linestring
        
        # Simplify
        coords_simplified = simplify_coords(linestring.coords)

        # If preserve_topology is True and the result is no line anymore, return original line
        if(preserve_topology is True
           and (coords_simplified is None or len(coords_simplified) < 2)):
            return linestring
        else:
            return sh_geom.LineString(coords_simplified)

    def simplify_coords(coords: list) -> List[Any]:
        # Determine the indexes of the coordinates to keep after simplification
        if algorithm is SimplifyAlgorithm.RAMER_DOUGLAS_PEUCKER:
            coords_simplify_idx = simplification.simplify_coords_idx(coords, tolerance)
        elif algorithm is SimplifyAlgorithm.VISVALINGAM_WHYATT:
            coords_simplify_idx = simplification.simplify_coords_vw_idx(coords, tolerance)
        elif algorithm is SimplifyAlgorithm.LANG:
            coords_simplify_idx = simplify_coords_lang_idx(coords, tolerance, lookahead=lookahead)    
        else:
            raise Exception(f"Unsupported algorithm: {algorithm}, supported: {SimplifyAlgorithm}")

        coords_on_border_idx = []
        if keep_points_on is not None:
            coords_gdf = gpd.GeoDataFrame(geometry=list(sh_geom.MultiPoint(coords)))
            coords_on_border_series = coords_gdf.intersects(keep_points_on)
            coords_on_border_idx = np.array(coords_on_border_series.index[coords_on_border_series]).tolist()

        # Extracts coordinates that need to be kept
        coords_to_keep = sorted(set(coords_simplify_idx + coords_on_border_idx))
        return np.array(coords)[coords_to_keep].tolist()
        
    # Loop over the rings, and simplify them one by one...
    # If the geometry is None, just return...
    if geometry is None:
        raise Exception("geom input paramerter should not be None")
    elif isinstance(geometry, sh_geom.Point):
        # Point cannot be simplified
        return geometry
    elif isinstance(geometry, sh_geom.MultiPoint):
        # MultiPoint cannot be simplified
        return geometry
    elif isinstance(geometry, sh_geom.LineString):
        result_geom = simplify_linestring(geometry)
    elif isinstance(geometry, sh_geom.Polygon):
        result_geom = simplify_polygon(geometry)
    elif isinstance(geometry, sh_geom.base.BaseMultipartGeometry):
        # If it is a multi-part, recursively call simplify for all parts. 
        simplified_geometries = []
        for geom in list(geometry):
            simplified_geometries.append(simplify_ext(geom, 
                    tolerance=tolerance, 
                    algorithm=algorithm, lookahead=lookahead, 
                    preserve_topology=preserve_topology, 
                    keep_points_on=keep_points_on))
        result_geom = collect(simplified_geometries)
    else:
        raise Exception(f"Unsupported geom_type: {geometry.geom_type}, {geometry}")

    return make_valid(result_geom)

def simplify_coords_lang(
        coords: Union[np.ndarray, list],
        tolerance: float,
        lookahead: int) -> Union[np.ndarray, list]:
    """
    Simplify a line using lang algorithm.

    Args:
        coords (Union[np.ndarray, list]): list of coordinates to be simplified.
        tolerance (float): distance tolerance to use.
        lookahead (int, optional): the number of points to consider for removing
            in a moving window. Defaults to 8.

    Returns:
        Return the coordinates kept after simplification. 
        If input coords is np.ndarray, returns np.ndarray, otherwise returns a list.
    """

    # Init variables 
    if isinstance(coords, np.ndarray):
        coords_arr = coords
    else:
        coords_arr = np.array(list(coords))

    # Determine the coordinates that need to be kept 
    coords_to_keep_idx = simplify_coords_lang(
            coords=coords_arr,
            tolerance=tolerance,
            lookahead=lookahead)
    coords_simplified_arr = coords_arr[coords_to_keep_idx]

    # If input was np.ndarray, return np.ndarray, otherwise list
    if isinstance(coords, np.ndarray):
        return coords_simplified_arr
    else:
        return coords_simplified_arr.tolist()

def simplify_coords_lang_idx(
        coords: Union[np.ndarray, list],
        tolerance: float,
        lookahead: int = 8) -> Union[np.ndarray, list]:
    """
    Simplify a line using lang algorithm and return the coordinate indexes to 
    be kept.

    Inspiration for the implementation came from:
        * https://github.com/giscan/Generalizer/blob/master/simplify.py
        * https://github.com/keszegrobert/polyline-simplification/blob/master/6.%20Lang.ipynb
        * https://web.archive.org/web/20171005193700/http://web.cs.sunyit.edu/~poissad/projects/Curve/about_algorithms/lang.php

    Args:
        coords (Union[np.ndarray, list]): list of coordinates to be simplified.
        tolerance (float): distance tolerance to use.
        lookahead (int, optional): the number of points to consider for removing
            in a moving window. Defaults to 8.

    Returns:
        Return the indexes of coordinates that need to be kept after 
        simplification. 
        If input coords is np.ndarray, returns np.ndarray, otherwise returns a list.  
    """
    
    def point_line_distance(point_x, point_y, line_x1, line_y1, line_x2, line_y2) -> float:
        denominator = math.sqrt((line_x2-line_x1)*(line_x2-line_x1)+(line_y2-line_y1)*(line_y2-line_y1))
        if denominator == 0:
            return float('Inf')
        else:
            numerator = abs((line_x2-line_x1)*(line_y1-point_y)-(line_x1-point_x)*(line_y2-line_y1)) 
            return numerator/denominator

    # Init variables 
    if isinstance(coords, np.ndarray):
        line_arr = coords
    else:
        line_arr = np.array(list(coords))

    # Prepare lookahead
    nb_points = len(line_arr)
    if lookahead == -1:
        window_size = nb_points-1
    else:
        window_size = min(lookahead, nb_points-1)

    #mask = np.ones(nb_points), dtype='bool')
    mask = np.zeros(nb_points, dtype='bool')
    mask[0] = True
    window_start = 0
    window_end = window_size
    
    # Apply simplification till the window_start arrives at the last point.
    ready = False
    while ready is False:
        # Check if all points between window_start and window_end are within  
        # tolerance distance to the line (window_start, window_end).
        all_points_in_tolerance = True
        for i in range(window_start+1, window_end):
            distance = point_line_distance(
                    line_arr[i, 0], line_arr[i, 1],
                    line_arr[window_start, 0], line_arr[window_start, 1],
                    line_arr[window_end, 0], line_arr[window_end, 1])
            # If distance is nan (= linepoint1 == linepoint2) or > tolerance
            if distance > tolerance:
                all_points_in_tolerance = False
                break

        # If not all the points are within the tolerance distance... 
        if not all_points_in_tolerance:
            # Move window_end to previous point, and try again
            window_end -= 1
        else:
            # All points are within the tolerance, so they can be masked
            mask[window_end] = True
            #mask[window_start+1:window_end-1] = False

            # Move the window forward
            window_start = window_end
            if window_start == nb_points - 1:
                ready = True
            window_end = window_start + window_size
            if window_end >= nb_points: 
                window_end = nb_points - 1
    
    # Prepare result: convert the mask to a list of indices of points to keep. 
    idx_to_keep_arr = mask.nonzero()[0]

    # If input was np.ndarray, return np.ndarray, otherwise list
    if isinstance(coords, np.ndarray):
        return idx_to_keep_arr
    else:
        return idx_to_keep_arr.tolist()



# -*- coding: utf-8 -*-
"""
Module containing utilities regarding the usage of ogr functionalities.
"""

import datetime
import enum
import logging
import os
from pathlib import Path
import sqlite3
from typing import Any, List

import pandas as pd

from geofileops.util.geometry_util import GeometryType
from geofileops import geofile
from .general_util import MissingRuntimeDependencyError

#-------------------------------------------------------------
# First define/init some general variables/constants
#-------------------------------------------------------------

# Get a logger...
logger = logging.getLogger(__name__)

class EmptyResultError(Exception):
    """
    Exception raised when the SQL statement disn't return any rows.

    Attributes:
        message (str): Exception message
    """
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

#-------------------------------------------------------------
# The real work
#-------------------------------------------------------------

def check_runtimedependencies():
    test_path = Path(__file__).resolve().parent / "test.gpkg"
    conn = sqlite3.connect(test_path)
    load_spatialite(conn)

class SqliteProfile(enum.Enum):
    DEFAULT=0
    SPEED=1

def create_new_spatialdb(
        path: Path, 
        crs_epsg: int = None):

    # Connect to sqlite
    conn = sqlite3.connect(path)
    sql = None
    try:
        with conn:
            load_spatialite(conn)
                    
            # Init file
            output_suffix_lower = path.suffix.lower()
            if output_suffix_lower == '.gpkg':
                sql = 'SELECT EnableGpkgMode();'
                #sql = 'SELECT EnableGpkgAmphibiousMode();'
                conn.execute(sql)
                # Remark: this only works on the main database!
                sql = 'SELECT gpkgCreateBaseTables();'
                conn.execute(sql)
                if crs_epsg is not None and crs_epsg not in [0, -1, 4326]:
                    sql = f"SELECT gpkgInsertEpsgSRID({crs_epsg})"
                    conn.execute(sql)
            elif output_suffix_lower == '.sqlite':
                sql = 'SELECT InitSpatialMetaData(1);'
                conn.execute(sql)
                if crs_epsg is not None and crs_epsg not in [0, -1, 4326]:
                    sql = f"SELECT InsertEpsgSrid({crs_epsg})"
                    conn.execute(sql)
            else:
                raise Exception(f"Unsupported output format: {output_suffix_lower}")

    except Exception as ex:
        raise Exception(f"Error creating spatial db {path}") from ex
    finally:
        conn.close()

def create_table_as_sql(
        input1_path: Path,
        input1_layer: str,
        input2_path: Path, 
        output_path: Path,
        sql_stmt: str,
        output_layer: str,
        output_geometrytype: GeometryType,
        append: bool = False,
        update: bool = False,
        create_spatial_index: bool = True,
        profile: SqliteProfile = SqliteProfile.DEFAULT):

    # Check input parameters
    if append is True or update is True:
        raise Exception('Not implemented')
    output_suffix_lower = input1_path.suffix.lower()
    if output_suffix_lower != input1_path.suffix.lower():
        raise Exception("Output and input1 paths don't have the same extension!")
    if(input2_path is not None 
       and output_suffix_lower != input2_path.suffix.lower()):
        raise Exception("Output and input2 paths don't have the same extension!")
        
    # Use crs epsg from input1_layer, if it has one
    input1_layerinfo = geofile.get_layerinfo(input1_path, input1_layer)
    crs_epsg = -1
    if input1_layerinfo.crs is not None and input1_layerinfo.crs.to_epsg() is not None: 
        crs_epsg = input1_layerinfo.crs.to_epsg()

    # If output file doesn't exist yet, create and init it
    if not output_path.exists():
        create_new_spatialdb(path=output_path, crs_epsg=crs_epsg)
        
    sql = None
    conn = sqlite3.connect(output_path, detect_types=sqlite3.PARSE_DECLTYPES)
    try:
        with conn:
            def to_string_for_sql(input) -> str:
                if input is None:
                    return 'NULL'
                else:
                    return str(input)

            # Connect to output database file so it is main, otherwise the 
            # gpkg... functions don't work
            # Remark: sql statements using knn only work if they are main, so they 
            # are executed with ogr, as the output needs to be main as well :-(. 
            output_databasename = 'main'
            load_spatialite(conn)

            if output_suffix_lower == '.gpkg':
                sql = 'SELECT EnableGpkgMode();'
                conn.execute(sql)

            # Set number of cache pages (1 page = 4096 bytes)
            conn.execute('PRAGMA cache_size = 10000;')
            conn.execute('PRAGMA temp_store = MEMORY;')
            
            # Use the sqlite profile specified
            if profile is SqliteProfile.SPEED:
                conn.execute(f"PRAGMA journal_mode = OFF;")
            
                # These pragma's increase speed
                conn.execute('PRAGMA locking_mode = EXCLUSIVE;')
                conn.execute('PRAGMA synchronous = OFF;')
                
                # Use memory mapped IO = much faster (max 30GB)
                conn.execute('PRAGMA mmap_size = 30000000000;')

            # If attach to input1
            input1_databasename = 'input1'
            sql = f"ATTACH DATABASE ? AS {input1_databasename}"
            dbSpec = (str(input1_path),)
            conn.execute(sql, dbSpec)

            # If input2 isn't the same database input1, attach to it
            if input2_path is not None:
                if input2_path == input1_path:
                    input2_databasename = input1_databasename
                else:
                    input2_databasename = 'input2'
                    sql = f"ATTACH DATABASE ? AS {input2_databasename}"
                    dbSpec = (str(input2_path),)
                    conn.execute(sql, dbSpec)    
            
            # Prepare sql statement
            sql_stmt = sql_stmt.format(
                    input1_databasename=input1_databasename,
                    input2_databasename=input2_databasename)

            ### Determine columns/datatypes to create the table ###
            # Create temp table to get the column names + general data types
            # + fetch one row to use it to determine geometrytype.
            sql = f'CREATE TEMPORARY TABLE tmp AS \nSELECT * FROM (\n{sql_stmt}\n)\nLIMIT 1;'
            conn.execute(sql)
            sql = 'PRAGMA TABLE_INFO(tmp)'
            cur = conn.execute(sql)
            tmpcolumns = cur.fetchall()
                        
            # Fetch one row to try to get more detailed data types
            sql = sql_stmt
            tmpdata = conn.execute(sql).fetchone()
            if tmpdata is None or len(tmpdata) == 0:
                # If no row was returned, stop
                raise EmptyResultError(f"Query didn't return any rows: {sql_stmt}")
            
            # Loop over all columns to determine the data type
            column_types = {}
            for column_index, column in enumerate(tmpcolumns):
                columnname = column[1]
                columntype = column[2]

                if columnname == 'geom':
                    # PRAGMA TABLE_INFO gives None as column type for a 
                    # geometry column. So if output_geometrytype not specified, 
                    # Use ST_GeometryType to get the type
                    # based on the data + apply to_multitype to be sure
                    if output_geometrytype is None:
                        sql = f'SELECT ST_GeometryType({columnname}) FROM tmp;'
                        column_geometrytypename = conn.execute(sql).fetchall()[0][0]
                        output_geometrytype = GeometryType[column_geometrytypename].to_multitype
                    column_types[columnname] = output_geometrytype.name
                else:
                    # If PRAGMA TABLE_INFO doesn't specify the datatype, 
                    # determine based on data
                    if columntype is None or columntype == '':
                        sql = f'SELECT typeof({columnname}) FROM tmp;'
                        column_types[columnname] = conn.execute(sql).fetchall()[0][0]
                    elif columntype == 'NUM':
                        # PRAGMA TABLE_INFO sometimes returns 'NUM', but 
                        # apparently this cannot be used in "CREATE TABLE" 
                        if isinstance(tmpdata[column_index], datetime.date): 
                            column_types[columnname] = 'DATE'
                        elif isinstance(tmpdata[column_index], datetime.datetime): 
                            column_types[columnname] = 'DATETIME'
                        else:
                            column_types[columnname] = 'DECIMAL'
                    else:
                        column_types[columnname] = columntype

            ### Now we can create the table ###
            # Create output table using the gpkgAddGeometryColumn() function
            # Problem: the spatialite function gpkgAddGeometryColumn() doesn't support 
            # layer names with special characters (eg. '-')... 
            # Solution: mimic the behaviour of gpkgAddGeometryColumn manually.
            # Create table without geom column
            ''' 
            columns_for_create = [f'"{columnname}" {column_types[columnname]}\n' for columnname in column_types if columnname != 'geom']
            sql = f'CREATE TABLE {output_databasename}."{output_layer}" ({", ".join(columns_for_create)})'
            conn.execute(sql)
            # Add geom column with gpkgAddGeometryColumn()
            # Remark: output_geometrytype.name should be detemined from data if needed, see above...
            sql = f"SELECT gpkgAddGeometryColumn('{output_layer}', 'geom', '{output_geometrytype.name}', 0, 0, {to_string_for_sql(crs_epsg)});"
            conn.execute(sql)
            '''
            columns_for_create = [f'"{columnname}" {column_types[columnname]}\n' for columnname in column_types]
            sql = f'CREATE TABLE {output_databasename}."{output_layer}" ({", ".join(columns_for_create)})'
            conn.execute(sql)

            # Add metadata 
            if output_suffix_lower == '.gpkg':
                # ~ mimic behaviour of gpkgAddGeometryColumn()
                sql = f"""
                        INSERT INTO {output_databasename}.gpkg_contents (
                                table_name, data_type, identifier, description, last_change, 
                                min_x, min_y, max_x, max_y, srs_id)
                            VALUES ('{output_layer}', 'features', NULL, '', DATETIME(),
                                NULL, NULL, NULL, NULL, {to_string_for_sql(crs_epsg)});"""
                conn.execute(sql)
                sql = f"""
                        INSERT INTO {output_databasename}.gpkg_geometry_columns (
                                table_name, column_name, geometry_type_name, srs_id, z, m)
                            VALUES ('{output_layer}', 'geom', '{output_geometrytype.name}', 
                                    {to_string_for_sql(crs_epsg)}, 0, 0);"""
                conn.execute(sql)
                
                # Now add geom triggers
                # Remark: this only works on the main database!
                sql = f"SELECT gpkgAddGeometryTriggers('{output_layer}', 'geom');"
                conn.execute(sql)
            elif output_suffix_lower == '.sqlite':
                # Create metadata for the geometry column
                sql = f"""
                        SELECT RecoverGeometryColumn(
                            '{output_layer}', 'geom',
                            {to_string_for_sql(crs_epsg)}, '{output_geometrytype.name}');"""
                conn.execute(sql)
            
            # Insert data using the sql statement specified
            columns_for_insert = [f'"{column[1]}"' for column in tmpcolumns]
            sql = f'INSERT INTO {output_databasename}."{output_layer}" ({", ".join(columns_for_insert)})\n{sql_stmt}' 
            conn.execute(sql)           
                    
            # Create spatial index if needed
            if create_spatial_index is True:
                sql = f"SELECT UpdateLayerStatistics('{output_layer}', 'geom');"
                conn.execute(sql)
                if output_suffix_lower == '.gpkg':
                    sql = f"SELECT gpkgAddSpatialIndex('{output_layer}', 'geom');"
                elif output_suffix_lower == '.sqlite':
                    sql = f"SELECT CreateSpatialIndex('{output_layer}', 'geom');"
                conn.execute(sql)
    
    except EmptyResultError as ex:
        logger.info(f"Query didn't return any rows: {sql_stmt}")
        if output_path.exists():
            output_path.unlink()
    except Exception as ex:
        raise Exception(f"Error executing {sql}") from ex
    finally:
        conn.close()
        
def execute_sql(
        path: Path,
        sql_stmt: str,
        use_spatialite: bool = True):

    # Connect to database file
    conn = sqlite3.connect(path, detect_types=sqlite3.PARSE_DECLTYPES)    
    sql = None

    try:
        with conn:
            if use_spatialite is True:
                load_spatialite(conn)
            if path.suffix.lower() == '.gpkg':
                sql = 'SELECT EnableGpkgMode();'
                conn.execute(sql)
            
            # Set number of cache pages (1 page = 4096 bytes)
            sql = 'PRAGMA cache_size = 10000;'
            conn.execute(sql)
            sql = 'PRAGMA temp_store = MEMORY;'
            conn.execute(sql)

            # Now actually run the sql
            sql = sql_stmt
            conn.execute(sql)

    except Exception as ex:
        raise Exception(f"Error executing {sql}") from ex
    finally:
        conn.close()

def execute_select_sql(
        path: Path,
        sql_stmt: str,
        use_spatialite: bool = True) -> List[Any]:

    # Connect to database file
    conn = sqlite3.connect(path, detect_types=sqlite3.PARSE_DECLTYPES)
    sql = None
    
    try:    
        if use_spatialite is True:
            load_spatialite(conn)
        if path.suffix.lower() == '.gpkg':
            sql = 'SELECT EnableGpkgMode();'
            conn.execute(sql)
        
        # Set number of cache pages (1 page = 4096 bytes)
        sql = 'PRAGMA cache_size = 10000;'
        conn.execute(sql)
        # Use memory mapped IO = much faster (max 30GB)
        conn.execute('PRAGMA mmap_size = 30000000000;')

        sql = sql_stmt
        return conn.execute(sql).fetchall()

    except Exception as ex:
        raise Exception(f"Error executing {sql}") from ex
    finally:
        conn.close()

def execute_select_sql_df(
        path: Path,
        sql_stmt: str,
        use_spatialite: bool = True) -> pd.DataFrame:

    # Connect to database file
    conn = sqlite3.connect(path, detect_types=sqlite3.PARSE_DECLTYPES)
    sql = None

    try:
        if use_spatialite is True:
            load_spatialite(conn)
        if path.suffix.lower() == '.gpkg':
            sql = 'SELECT EnableGpkgMode();'
            conn.execute(sql)
        
        # Set number of cache pages (1 page = 4096 bytes)
        sql = 'PRAGMA cache_size = 10000;'
        conn.execute(sql)
        # Use memory mapped IO = much faster (max 30GB)
        conn.execute('PRAGMA mmap_size = 30000000000;')

        sql = sql_stmt
        return pd.read_sql(sql, conn)

    except Exception as ex:
        raise Exception(f"Error executing {sql}") from ex
    finally:
        conn.close()

def load_spatialite(conn):
    """
    Load mod_spatialite for an existing sqlite connection.

    Args:
        conn ([type]): Sqlite connection
    """
    conn.enable_load_extension(True)
    try:
        conn.load_extension('mod_spatialite')
    except Exception as ex:
        mod_spatialite_dir = os.environ.get('MOD_SPATIALITE_DIR')
        if mod_spatialite_dir is not None:
            if f"{mod_spatialite_dir};" not in os.environ['PATH']:
                os.environ['PATH'] = f"{mod_spatialite_dir};{os.environ['PATH']}"
            try:
                conn.load_extension('mod_spatialite')
            except Exception as ex:
                raise MissingRuntimeDependencyError(f"Error trying to load mod_spatialite with MOD_SPATIALITE_DIR: {mod_spatialite_dir}") from ex
        else:
            raise MissingRuntimeDependencyError("Error trying to load mod_spatialite. You can specify the location of mod_spatialite using the MOD_SPATIALITE_DIR environment variable") from ex
    
if __name__ == '__main__':
    None
